# RCAIDE/Methods/Noise/Correlation_Buildup/Airframe/airframe_noise.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE Imports
import  RCAIDE
from RCAIDE.Framework.Core                      import Data , Units 
from .clean_wing_noise             import clean_wing_noise
from .landing_gear_noise           import landing_gear_noise
from .leading_edge_slat_noise      import leading_edge_slat_noise
from .trailing_edge_flap_noise     import trailing_edge_flap_noise 
from RCAIDE.Library.Methods.Noise.Metrics  import PNL_noise_metric
from RCAIDE.Library.Methods.Noise.Metrics  import EPNL_noise_metric 
from RCAIDE.Library.Methods.Noise.Metrics  import A_weighting_metric 
from RCAIDE.Library.Methods.Noise.Common   import noise_tone_correction
from RCAIDE.Library.Methods.Noise.Common   import atmospheric_attenuation
from RCAIDE.Library.Methods.Noise.Common   import SPL_arithmetic 

# python imports 
import numpy as np

# ----------------------------------------------------------------------
#  Airframe Noise 
# ----------------------------------------------------------------------
def airframe_noise(microphone_locations,segment,config,settings):  
    """ This computes the noise from different sources of the airframe for a given vehicle for a constant altitude flight. 

    Assumptions:
        Correlation based 
 
    Source:
        Fink, Martin R. Airframe noise prediction method. No. UTRC/R77-912607-11. UNITED 
        TECHNOLOGIES RESEARCH CENTER EAST HARTFORD CT, 1977.  
               
    Inputs:
        vehicle	 - RCAIDE type vehicle

        includes these fields:
            S                          - Wing Area
            bw                         - Wing Span
            Sht                        - Horizontal tail area
            bht                        - Horizontal tail span
            Svt                        - Vertical tail area
            bvt                        - Vertical tail span
            deltaf                     - Flap deflection
            Sf                         - Flap area
            cf                         - Flap chord
            slots                      - Number of slots (Flap type)
            Dp                         - Main landing gear tyre diameter
            Hp                         - Main lading gear strut length
            Dn                         - Nose landing gear tyre diameter
            Hn                         - Nose landing gear strut length
            wheels                     - Number of wheels 
            
        noise segment - flight path data, containing:
            distance_vector             - distance from the source location to observer
            angle                       - polar angle from the source to the observer
            phi                         - azimuthal angle from the source to the observer


    Outputs: One Third Octave Band SPL [dB]
        SPL_wing                        - Sound Pressure Level of the clean wing
        SPLht                           - Sound Pressure Level of the horizontal tail
        SPLvt                           - Sound Pressure Level of the vertical tail
        SPL_flap                        - Sound Pressure Level of the flaps trailing edge
        SPL_slat                        - Sound Pressure Level of the slat leading edge
        SPL_main_landing_gear           - Sound Pressure Level og the main landing gear
        SPL_nose_landing_gear           - Sound Pressure Level of the nose landing gear

    Properties Used:
        N/A      
        
    """  
    # Unpack 
    wing     = config.wings
    flap     = wing.main_wing.control_surfaces.flap 
    Sw       = wing.main_wing.areas.reference  / (Units.ft)**2              # wing area, sq.ft
    bw       = wing.main_wing.spans.projected / Units.ft                    # wing span, ft
    Sht      = wing.horizontal_stabilizer.areas.reference / (Units.ft)**2   # horizontal tail area, sq.ft
    bht      = wing.horizontal_stabilizer.spans.projected / Units.ft        # horizontal tail span, ft
    Svt      = wing.vertical_stabilizer.areas.reference / (Units.ft)**2     # vertical tail area, sq.ft
    bvt      = wing.vertical_stabilizer.spans.projected  / Units.ft         # vertical tail span, ft
    deltaf   = flap.deflection                                              # flap delection, rad
    Sf       = flap.area  / (Units.ft)**2                                   # flap area, sq.ft        
    cf       = flap.chord_dimensional  / Units.ft                           # flap chord, ft

    Dp           = 0
    Dn           = 0
    main_wheels  = 0
    main_units   = 0
    Hp           = 0
    Hn           = 0 
    nose_wheels  = 0 
    for landing_gear in  config.landing_gears:
        if isinstance(landing_gear,RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear):
            Dp            = landing_gear.tire_diameter  / Units.ft           # MLG tyre diameter, ft
            Dn            = landing_gear.strut_length   / Units.ft           # NLG tyre diameter, ft
            main_wheels   = landing_gear.wheels                          # Number of wheels   
            gear_extended = landing_gear.gear_extended                   # Gear up or gear down 
            main_units    = landing_gear.units                           # Number of main units
        elif isinstance(landing_gear,RCAIDE.Library.Components.Landing_Gear.Nose_Landing_Gear):
            Hp            = landing_gear.tire_diameter  / Units.ft           # MLG strut length, ft
            Hn            = landing_gear.strut_length   / Units.ft           # NLG strut length, ft
            gear_extended = landing_gear.gear_extended                   # Gear up or gear down 
            nose_wheels   = landing_gear.wheels                          # Number of wheels   
    
    
    velocity     = segment.conditions.freestream.velocity                  # aircraft velocity  
    noise_time   = segment.conditions.frames.inertial.time[:,0]             # time discretization 

    # determining flap slot number
    if wing.main_wing.control_surfaces.flap.configuration_type   == 'single_slotted':
        slots = 1
    elif wing.main_wing.control_surfaces.flap.configuration_type == 'double_slotted':
        slots = 2
    elif wing.main_wing.control_surfaces.flap.configuration_type == 'triple_slotted':
        slots = 3
        
    
    # Generate array with the One Third Octave Band Center Frequencies
    frequency = settings.center_frequencies[5:]  
    num_f     = len(frequency) 
    n_cpts    = len(noise_time)  
    n_mic     = len(microphone_locations)        

    # Geometric information from the source to observer position  
    distance_vector     = np.linalg.norm(microphone_locations,axis = 1)
    thetas              = np.tile(np.arctan2(microphone_locations[:, 1], microphone_locations[:, 0])[None,:],(n_cpts,1)) 
    phis                = np.tile(np.arctan2(np.sqrt(np.square(microphone_locations[:, 0]) + np.square(microphone_locations[:, 1])),  microphone_locations[:, 2])[None,:],(n_cpts,1))  
    viscosity           = segment.conditions.freestream.dynamic_viscosity[:,0]*Units.ft*Units.ft  
    M                   = segment.conditions.freestream.mach_number
     
    SPL_wing_history              = np.zeros((n_cpts,n_mic,num_f))
    SPLht_history                 = np.zeros((n_cpts,n_mic,num_f))
    SPLvt_history                 = np.zeros((n_cpts,n_mic,num_f))
    SPL_flap_history              = np.zeros((n_cpts,n_mic,num_f))
    SPL_slat_history              = np.zeros((n_cpts,n_mic,num_f))
    SPL_main_landing_gear_history = np.zeros((n_cpts,n_mic,num_f))
    SPL_nose_landing_gear_history = np.zeros((n_cpts,n_mic,num_f))
    SPL_total_history             = np.zeros((n_cpts,n_mic,num_f)) 
    SPLt_dBA_history              = np.zeros((n_cpts,n_mic,num_f))  
    SPLt_dBA_max                  = np.zeros((n_cpts,n_mic))    
    
    # START LOOP FOR EACH POSITION OF AIRCRAFT   
    for i in range(n_cpts):
        for j in range(n_mic): 
            
            # Emission angle
            theta = thetas[i,j]
            phi   = phis[i,j]
            
            # Distance from airplane to observer, evaluated at retarded time
            distance = distance_vector[j]    
           
            # Atmospheric attenuation
            delta_atmo=atmospheric_attenuation(np.array([distance]),frequency)[:,0]
    
            # Call each noise source model
            SPL_wing = clean_wing_noise(Sw,bw,0,1,velocity[i,0],viscosity[i],M[i],phi,theta,distance,frequency) - delta_atmo    #Wing Noise
            SPLht    = clean_wing_noise(Sht,bht,0,1,velocity[i,0],viscosity[i],M[i],phi,theta,distance,frequency)  -delta_atmo    #Horizontal Tail Noise
            SPLvt    = clean_wing_noise(Svt,bvt,0,0,velocity[i,0],viscosity[i],M[i],phi,theta,distance,frequency)  -delta_atmo    #Vertical Tail Noise
     
            SPL_slat = leading_edge_slat_noise(SPL_wing,Sw,bw,velocity[i,0],viscosity[i],M[i],phi,theta,distance,frequency) -delta_atmo        #Slat leading edge
     
            if (deltaf==0):
                SPL_flap = np.zeros(num_f)
            else:
                SPL_flap = trailing_edge_flap_noise(Sf,cf,deltaf,slots,velocity[i,0],M[i],phi,theta,distance,frequency) - delta_atmo #Trailing Edge Flaps Noise
     
            if gear_extended == False:  
                SPL_main_landing_gear = np.zeros(num_f)
                SPL_nose_landing_gear = np.zeros(num_f)
            else:
                SPL_main_landing_gear = landing_gear_noise(Dp,Hp,main_wheels,M[i],velocity[i,0],phi,theta,distance,frequency)  - delta_atmo     #Main Landing Gear Noise
                SPL_nose_landing_gear = landing_gear_noise(Dn,Hn,nose_wheels,M[i],velocity[i,0],phi,theta,distance,frequency)  - delta_atmo     #Nose Landing Gear Noise
            if main_units>1: # Incoherent summation of each main landing gear unit
                SPL_main_landing_gear = SPL_main_landing_gear+3*(main_units-1) 
     
            # Total Airframe Noise
            SPL_total = 10.*np.log10(10.0**(0.1*SPL_wing)+10.0**(0.1*SPLht)+10**(0.1*SPL_flap)+ \
                 10.0**(0.1*SPL_slat)+10.0**(0.1*SPL_main_landing_gear)+10.0**(0.1*SPL_nose_landing_gear))
                
            SPL_total_history[i,j,:]             = SPL_total 
            SPL_wing_history[i,j,:]              = SPL_wing 
            SPLvt_history[i,j,:]                 = SPLvt 
            SPLht_history[i,j,:]                 = SPLht 
            SPL_flap_history[i,j,:]              = SPL_flap 
            SPL_slat_history[i,j,:]              = SPL_slat 
            SPL_nose_landing_gear_history[i,j,:] = SPL_nose_landing_gear 
            SPL_main_landing_gear_history[i,j,:] = SPL_main_landing_gear  
            
            # Calculation of dBA based on the sound pressure time history 
            SPLt_dBA_history[i,j,:] = A_weighting_metric(SPL_total,frequency)
            SPLt_dBA_max[i,j]       = np.max(A_weighting_metric(SPL_total,frequency))    
              
    # Calculation of the Perceived Noise Level EPNL based on the sound time history 
    PNL_total             = PNL_noise_metric(SPL_total_history)
    PNL_wing              = PNL_noise_metric(SPL_wing_history)
    PNL_ht                = PNL_noise_metric(SPLht_history)
    PNL_vt                = PNL_noise_metric(SPLvt_history)
    PNL_nose_landing_gear = PNL_noise_metric(SPL_nose_landing_gear_history)
    PNL_main_landing_gear = PNL_noise_metric(SPL_main_landing_gear_history)
    PNL_slat              = PNL_noise_metric(SPL_slat_history)
    PNL_flap              = PNL_noise_metric(SPL_flap_history)
     
    # Calculation of the tones corrections on the SPL for each component and total
    tone_correction_total             = noise_tone_correction(SPL_total_history) 
    tone_correction_wing              = noise_tone_correction(SPL_wing_history)
    tone_correction_ht                = noise_tone_correction(SPLht_history)
    tone_correction_vt                = noise_tone_correction(SPLvt_history)
    tone_correction_flap              = noise_tone_correction(SPL_flap_history)
    tone_correction_slat              = noise_tone_correction(SPL_slat_history)
    tone_correction_nose_landing_gear = noise_tone_correction(SPL_nose_landing_gear_history)
    tone_correction_main_landing_gear = noise_tone_correction(SPL_main_landing_gear_history)
    
    # Calculation of the PLNT for each component and total
    PNLT_total             = PNL_total+tone_correction_total
    PNLT_wing              = PNL_wing+tone_correction_wing
    PNLT_ht                = PNL_ht+tone_correction_ht
    PNLT_vt                = PNL_vt+tone_correction_vt
    PNLT_nose_landing_gear = PNL_nose_landing_gear+tone_correction_nose_landing_gear
    PNLT_main_landing_gear = PNL_main_landing_gear+tone_correction_main_landing_gear
    PNLT_slat              = PNL_slat+tone_correction_slat
    PNLT_flap              = PNL_flap+tone_correction_flap
    
    #Calculation of the EPNL for each component and total
    EPNL_total             = EPNL_noise_metric(PNLT_total)
    EPNL_wing              = EPNL_noise_metric(PNLT_wing)
    EPNL_ht                = EPNL_noise_metric(PNLT_ht)
    EPNL_vt                = EPNL_noise_metric(PNLT_vt)    
    EPNL_nose_landing_gear = EPNL_noise_metric(PNLT_nose_landing_gear)
    EPNL_main_landing_gear = EPNL_noise_metric(PNLT_main_landing_gear)
    EPNL_slat              = EPNL_noise_metric(PNLT_slat)
    EPNL_flap              = EPNL_noise_metric(PNLT_flap) 
    
    # Pack Airframe Noise 
    airframe_noise                        = Data()
    airframe_noise.EPNL_total             = EPNL_total
    airframe_noise.EPNL_wing              = EPNL_wing              
    airframe_noise.EPNL_ht                = EPNL_ht                
    airframe_noise.EPNL_vt                = EPNL_vt                
    airframe_noise.EPNL_nose_landing_gear = EPNL_nose_landing_gear 
    airframe_noise.EPNL_main_landing_gear = EPNL_main_landing_gear 
    airframe_noise.EPNL_slat              = EPNL_slat              
    airframe_noise.EPNL_flap              = EPNL_flap    
    airframe_noise.SPL                    = SPL_arithmetic(SPL_total_history, sum_axis= 2)
    airframe_noise.SPL_1_3_spectrum       = SPL_total_history
    airframe_noise.SPL_dBA                = SPL_arithmetic(np.atleast_2d(SPLt_dBA_history), sum_axis= 2) 
    airframe_noise.noise_time             = noise_time 
    return airframe_noise



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/General_Aviation/compute_landing_gear_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core import  Units ,  Data  

# ----------------------------------------------------------------------------------------------------------------------
# Main Wing Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_landing_gear_weight(landing_weight, Nult, strut_length_main, strut_length_nose):
    """ 
        Calculate the weight of the landing gear

        Source: Raymer- Aircraft Design: a Conceptual Approach (pg 460 in 4th edition)
        
        Inputs:
            Nult - ultimate landing load factor
            landing_weight- landing weight of the aircraft [kilograms]
           
        Outputs:
            weight - weight of the landing gear            [kilograms]
            
        Assumptions:
            calculating the landing gear weight based on the landing weight, load factor, and strut length 
    """ 

    #unpack
    W_l = landing_weight/Units.lbs
    l_n = strut_length_nose/Units.inches
    l_m = strut_length_main/Units.inches
    main_weight = .095*((Nult*W_l)**.768)*(l_m/12.)**.409
    nose_weight = .125*((Nult*W_l)**.566)*(l_n/12.)**.845

    # pack outputs
    output      = Data()
    output.main = main_weight*Units.lbs
    output.nose = nose_weight*Units.lbs

    return output


# RCAIDE/Library/Methods/Aerodynamics/Athena_Vortex_Lattice/write_run_cases.py
#
# Created: Oct 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
from RCAIDE.Library.Methods.Aerodynamics.Athena_Vortex_Lattice.purge_files       import purge_files
from RCAIDE.Library.Components.Wings.Control_Surfaces import Aileron , Elevator , Slat , Flap , Rudder  

# ----------------------------------------------------------------------------------------------------------------------
#  write_run_cases
# ---------------------------------------------------------------------------------------------------------------------- 
def write_run_cases(avl_object,trim_aircraft):
    """ This function writes the run cases used in the AVL batch analysis

    Assumptions:
        None
        
    Source:
        Drela, M. and Youngren, H., AVL, http://web.mit.edu/drela/Public/web/avl
   
    Inputs:
        avl_object.current_status.batch_file                    [-]
        avl_object.vehicle.mass_properties.center_of_gravity   [meters]
   
    Outputs:
        None
   
    Properties Used:
        N/A
    """       
    

    # unpack avl_inputs
    aircraft       = avl_object.vehicle
    batch_filename = avl_object.current_status.batch_file

    base_case_text = \
'''

 ---------------------------------------------
 Run case  {0}:   {1}

 alpha        ->  {2}       =   {3}        
 beta         ->  beta        =   {4}
 pb/2V        ->  pb/2V       =   {23}
 qc/2V        ->  qc/2V       =   {24}
 rb/2V        ->  rb/2V       =   0.00000
{5}
 alpha     =   {6}
 beta      =   0.00000     deg
 pb/2V     =   {25}
 qc/2V     =   {26}
 rb/2V     =   0.00000
 CL        =   {7}                        
 CDo       =   {8}
 bank      =   0.00000     deg
 elevation =   0.00000     deg
 heading   =   0.00000     deg
 Mach      =   {9}
 velocity  =   {10}     m/s               
 density   =   {11}     kg/m^3
 grav.acc. =   {12}     m/s^2
 turn_rad. =   0.00000     m
 load_fac. =   0.00000
 X_cg      =   {13}     m
 Y_cg      =   {14}     m
 Z_cg      =   {15}     m
 mass      =   {16}     kg
 Ixx       =   {17}     kg-m^2
 Iyy       =   {18}     kg-m^2
 Izz       =   {19}     kg-m^2
 Ixy       =   {20}     kg-m^2
 Iyz       =   {21}     kg-m^2
 Izx       =   {22}     kg-m^2
 visc CL_a =   0.00000
 visc CL_u =   0.00000
 visc CM_a =   0.00000
 visc CM_u =   0.00000

'''#{4} is a set of control surface inputs that will vary depending on the control surface configuration

    # Open the vehicle file after purging if it already exists
    purge_files([batch_filename]) 
    with open(batch_filename,'w') as runcases:
        # extract C.G. coordinates and moment of intertia tensor

        x_cg = aircraft.mass_properties.center_of_gravity[0][0]
        y_cg = aircraft.mass_properties.center_of_gravity[0][1]
        z_cg = aircraft.mass_properties.center_of_gravity[0][2]
        mass = aircraft.mass_properties.mass
        moments_of_inertia = aircraft.mass_properties.moments_of_inertia.tensor
        Ixx  = moments_of_inertia[0][0]
        Iyy  = moments_of_inertia[1][1]
        Izz  = moments_of_inertia[2][2]
        Ixy  = moments_of_inertia[0][1]
        Iyz  = moments_of_inertia[1][2]
        Izx  = moments_of_inertia[2][0]

        for case in avl_object.current_status.cases:
            # extract flight conditions 
            index = case.index
            name  = case.tag
            CL    = case.conditions.aerodynamics.coefficients.lift.total
            CDp   = 0.
            AoA   = round(case.conditions.aerodynamics.angles.alpha,4)
            beta  = round(case.conditions.aerodynamics.angles.beta,4)
            pb_2V = round(case.conditions.static_stability.coefficients.roll,4)
            qc_2V = round(case.conditions.static_stability.coefficients.pitch,4)
            mach  = round(case.conditions.freestream.mach,4)
            vel   = round(case.conditions.freestream.velocity,4)
            rho   = round(case.conditions.freestream.density,4)
            g     = case.conditions.freestream.gravitational_acceleration
            
            if trim_aircraft == False: # this flag sets up a trim analysis if one is declared by the boolean "trim_aircraft"
                controls_text = ''  
                if CL is not None: # if flight lift coefficient is specified without trim, the appropriate fields are filled 
                    toggle_idx = 'CL   '
                    toggle_val = round(CL,4)
                    alpha_val  = '0.00000     deg'
                    CL_val     = '0.00000'
                else: # if angle of attack is specified without trim, the appropriate fields are filled 
                    toggle_idx = 'alpha'
                    toggle_val = AoA
                    alpha_val  = '0.00000     deg'
                    CL_val     = '0.00000'
                if case.stability_and_control.number_of_control_surfaces != 0 :
                    # write control surface text in .run file if there is any
                    controls = make_controls_case_text(case.stability_and_control.control_surface_names,avl_object.vehicle)
                    controls_text = ''.join(controls)
 
            elif trim_aircraft: # trim is specified  
                if CL is not None:  # if flight lift coefficient is specified with trim, the appropriate fields are filled with the trim CL
                    toggle_idx = 'CL'
                    toggle_val = round(CL,4)
                    alpha_val  = '0.00000     deg'
                    CL_val     = CL
                else: # if angle of attack is specified with trim, the appropriate fields are filled with the trim AoA
                    toggle_idx = 'alpha'
                    toggle_val = AoA
                    alpha_val  = AoA 
                    CL_val     = '0.00000'
                
                controls = []
                if case.stability_and_control.number_of_control_surfaces != 0 :
                    # write control surface text in .run file if there is any
                    controls = make_controls_case_text(case.stability_and_control.control_surface_names,avl_object.vehicle)
                controls_text = ''.join(controls)
                
            # write the .run file using template and the extracted vehicle properties and flight condition
            case_text = base_case_text.format(index,name,toggle_idx,toggle_val,beta,controls_text,alpha_val, CL_val,CDp,
                                              mach,vel,rho,g,x_cg,y_cg,z_cg,mass,Ixx,Iyy,Izz,Ixy,Iyz,Izx,pb_2V,qc_2V,pb_2V,qc_2V) 
            runcases.write(case_text)

    return

def make_controls_case_text(cs_names,avl_aircraft):
    """ This function writes the text of the control surfaces in the AVL batch analysis.
    This tells AVL what control surface you want use to control a particular response.
    """ 
    control_surface_text = []
    
    for wing in avl_aircraft.wings: 
        for ctrl_surf in wing.control_surfaces: 
            if (type(ctrl_surf) ==  Slat):
                pass
            else:
                # template for control surface 
                base_control_cond_text = ' {0}->  {1}=   0.00000\n' 
                # This condition block assigns the control surface to a particular stabiltiy response 
                if (type(ctrl_surf) ==  Flap):         # if control surface function is 'flap', specify to AVL that it is a flap
                    ctrl_name = "{:<13}".format(ctrl_surf.tag)
                    variable  = 'flap        '
                elif (type(ctrl_surf) ==  Aileron):    # if control surface funcion is 'aileron', specify to AVL that this is used to produce no roll moment
                    ctrl_name = "{:<13}".format(ctrl_surf.tag)
                    variable  = 'Cl roll mom '
                elif (type(ctrl_surf) ==  Elevator):   # if control surface function is 'elevator', specify to AVL that this is used to produce no pitch moment
                    ctrl_name = "{:<13}".format(ctrl_surf.tag)
                    variable  = 'Cm pitchmom '
                elif (type(ctrl_surf) ==  Rudder):     # if control surface function is 'rudder', specify to AVL that this is used to produce no yaw moment
                    ctrl_name = "{:<13}".format(ctrl_surf.tag)
                    variable  = 'Cn yaw  mom '
                    
                # write the control surface functionality into template 
                text = base_control_cond_text.format(ctrl_name,variable)
            
                # append text 
                control_surface_text.append(text)
    return control_surface_text



# RCAIDE/Energy/Networks/Electric.py
# 
# Created:  Jul 2023, M. Clarke
# Modified: Sep 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports 
import RCAIDE  
from RCAIDE.Framework.Mission.Common                      import Residuals
from RCAIDE.Library.Mission.Common.Unpack_Unknowns.energy import unknowns
from .Network                                             import Network              
from RCAIDE.Library.Methods.Propulsors.Common.compute_avionics_power_draw import compute_avionics_power_draw
from RCAIDE.Library.Methods.Propulsors.Common.compute_payload_power_draw  import compute_payload_power_draw

# Python imports
import  numpy as  np

# ----------------------------------------------------------------------------------------------------------------------
#  Electric
# ----------------------------------------------------------------------------------------------------------------------  
class Electric(Network):
    """ A network comprising battery pack(s) to power rotors using electric motors via a bus.
        Electronic speed controllers, thermal management system, avionics, and other eletric 
        power systes paylaods are also modelled. Rotors and motors are arranged into groups,
        called propulsor groups, to siginify how they are connected in the network.
        The network also takes into consideration thermal management components that are
        connected to a coolant line.
        The network adds additional unknowns and residuals to the mission to determinge 
        the torque matching between motors and rotors in each propulsor group.


        Assumptions:
        The y axis rotation is used for rotating the rotor about the Y-axis for tilt rotors and tiltwings

        Source:
        None
    """  
    def __defaults__(self):
        """ This sets the default values for the network to function.

            Assumptions:
            None

            Source:
            N/A

            Inputs:
            None

            Outputs:
            None

            Properties Used:
            N/A
        """         

        self.tag                          = 'electric'
        self.system_voltage               = None   
        self.reverse_thrust               = False
        self.wing_mounted                 = True        

    # manage process with a driver function
    def evaluate(network,state,center_of_gravity):
        """ Calculate thrust given the current state of the vehicle

            Assumptions:
            Caps the throttle at 110% and linearly interpolates thrust off that

            Source:
            N/A

            Inputs:
            state [state()]

            Outputs:
            results.thrust_force_vector         [newtons]
            results.vehicle_mass_rate           [kg/s]
            conditions.energy                   [-]    

            Properties Used:
            Defaulted values
        """          

        # unpack   
        conditions      = state.conditions 
        busses          = network.busses
        coolant_lines   = network.coolant_lines
        total_thrust    = 0. * state.ones_row(3) 
        total_power     = 0. * state.ones_row(1) 
        total_moment    = 0. * state.ones_row(3)  
        reverse_thrust  = network.reverse_thrust

        for bus in busses:
            T               = 0. * state.ones_row(1) 
            total_power     = 0. * state.ones_row(1) 
            M               = 0. * state.ones_row(1)  
            avionics              = bus.avionics
            payload               = bus.payload  

            # Avionics Power Consumtion
            avionics_conditions = state.conditions.energy[bus.tag][avionics.tag]
            compute_avionics_power_draw(avionics,avionics_conditions,conditions)

            # Payload Power
            payload_conditions = state.conditions.energy[bus.tag][payload.tag]
            compute_payload_power_draw(payload,payload_conditions,conditions)

            # Bus Voltage 
            bus_voltage = bus.voltage * state.ones_row(1)

            if conditions.energy.recharging:
                avionics_power         = (avionics_conditions.power*bus.power_split_ratio)* state.ones_row(1)
                payload_power          = (payload_conditions.power*bus.power_split_ratio)* state.ones_row(1)            
                total_esc_power        = 0 * state.ones_row(1)
                bus.charging_current   = bus.nominal_capacity * bus.charging_c_rate 
                charging_power         = (bus.charging_current*bus_voltage*bus.power_split_ratio)

                # append bus outputs to battery
                bus_conditions                    = state.conditions.energy[bus.tag]
                bus_conditions.power_draw         = ((avionics_power + payload_power + total_esc_power) - charging_power)/bus.efficiency
                bus_conditions.current_draw       = -bus_conditions.power_draw/bus.voltage

            else:       
                # compute energy consumption of each battery on bus 
                stored_results_flag  = False
                stored_propulsor_tag = None 
                for propulsor_group in bus.assigned_propulsors:
                    for propulsor_tag in propulsor_group:
                        propulsor =  network.propulsors[propulsor_tag]
                        if propulsor.active and bus.active:       
                            if network.identical_propulsors == False:
                                # run analysis  
                                T,M,P,stored_results_flag,stored_propulsor_tag = propulsor.compute_performance(state,bus_voltage,center_of_gravity)
                            else:             
                                if stored_results_flag == False: 
                                    # run propulsor analysis 
                                    T,M,P,stored_results_flag,stored_propulsor_tag = propulsor.compute_performance(state,bus_voltage,center_of_gravity)
                                else:
                                    # use previous propulsor results 
                                    T,M,P = propulsor.reuse_stored_data(state,network,stored_propulsor_tag,center_of_gravity)
    
                            total_thrust += T   
                            total_moment += M   
                            total_power  += P 

                # compute power from each componemnt 
                avionics_power  = (avionics_conditions.power*bus.power_split_ratio)* state.ones_row(1) 
                payload_power   = (payload_conditions.power*bus.power_split_ratio)* state.ones_row(1)   
                charging_power  = (state.conditions.energy[bus.tag].regenerative_power*bus_voltage*bus.power_split_ratio) 
                total_esc_power = total_power*bus.power_split_ratio  

                # append bus outputs to battery 
                bus_conditions                    = state.conditions.energy[bus.tag]
                bus_conditions.power_draw        += ((avionics_power + payload_power + total_esc_power) - charging_power)/bus.efficiency
                bus_conditions.current_draw       = bus_conditions.power_draw/bus_voltage


        time               = state.conditions.frames.inertial.time[:,0] 
        delta_t            = np.diff(time)
        for t_idx in range(state.numerics.number_of_control_points):    
            for bus in  busses:
                stored_results_flag  = False
                stored_battery_tag   = None                          
                for battery_module in  bus.battery_modules:                   
                    if bus.identical_battery_modules == False:
                        # run analysis  
                        stored_results_flag, stored_battery_tag =  battery_module.energy_calc(state,bus,coolant_lines, t_idx, delta_t)
                    else:             
                        if stored_results_flag == False: 
                            # run battery analysis 
                            stored_results_flag, stored_battery_tag  =  battery_module.energy_calc(state,bus,coolant_lines, t_idx, delta_t)
                        else:
                            # use previous battery results 
                            battery_module.reuse_stored_data(state,bus,coolant_lines, t_idx, delta_t,stored_results_flag, stored_battery_tag)
                bus.compute_distributor_conditions(state,t_idx, delta_t)
                
                # Thermal Management Calculations                    
                for coolant_line in  coolant_lines:
                    if t_idx != state.numerics.number_of_control_points-1: 
                        for heat_exchanger in coolant_line.heat_exchangers: 
                            heat_exchanger.compute_heat_exchanger_performance(state,bus,coolant_line,delta_t[t_idx],t_idx) 
                        for reservoir in coolant_line.reservoirs:   
                            reservoir.compute_reservior_coolant_temperature(state,coolant_line,delta_t[t_idx],t_idx)
        
        if reverse_thrust ==  True:
            total_thrust =  total_thrust * -1     
            total_moment =  total_moment * -1                    
        conditions.energy.thrust_force_vector  = total_thrust
        conditions.energy.power                = total_power 
        conditions.energy.thrust_moment_vector = total_moment
        conditions.energy.vehicle_mass_rate    = state.ones_row(1)*0.0  

        return




    def unpack_unknowns(self,segment):
        """ This adds additional unknowns which are unpacked from the mission solver and send to the network.

            Assumptions:
            None

            Source:
            N/A

            Inputs:
            unknowns specific to the rotors                        [None] 
            unknowns specific to the battery cell                  

            Outputs:
            state.conditions.energy.bus.rotor.power_coefficient(s) [None] 
            conditions specific to the battery cell 

            Properties Used:
            N/A
        """
 
        unknowns(segment)
         
        for network in segment.analyses.energy.vehicle.networks:
            for bus_i, bus in enumerate(network.busses):    
                if bus.active:
                    for propulsor_group in  bus.assigned_propulsors:
                        propulsor = network.propulsors[propulsor_group[0]]
                        propulsor.unpack_propulsor_unknowns(segment) 
        return     

    def residuals(self,segment):
        """ This packs the residuals to be sent to the mission solver.

           Assumptions:
           None

           Source:
           N/A

           Inputs:
           state.conditions.energy:
               motor(s).torque                      [N-m]
               rotor(s).torque                      [N-m] 
           residuals soecific to the battery cell   

           Outputs:
           residuals specific to battery cell and network

           Properties Used: 
           N/A
        """
              
        for network in segment.analyses.energy.vehicle.networks:
            for bus_i, bus in enumerate(network.busses):    
                if bus.active:
                    for propulsor_group in  bus.assigned_propulsors:
                        propulsor =  network.propulsors[propulsor_group[0]]
                        propulsor.pack_propulsor_residuals(segment)   
        return
    
    def add_unknowns_and_residuals_to_segment(self, segment):
        """ This function sets up the information that the mission needs to run a mission segment using this network

            Assumptions:
            None

            Source:
            N/A

            Inputs:
            segment
            estimated_battery_voltages                                    [v]
            estimated_rotor_power_coefficients                            [float]s
            estimated_battery_cell_temperature                            [Kelvin]
            estimated_battery_state_of_charges                            [unitless]
            estimated_battery_cell_currents                               [Amperes]

            Outputs: 
            segment

            Properties Used:
            N/A
        """                
        segment.state.residuals.network = Residuals()
        
        for network in segment.analyses.energy.vehicle.networks:
            for p_i, propulsor in enumerate(network.propulsors): 
                propulsor.append_operating_conditions(segment)           
    
                for tag, propulsor_item in  propulsor.items():  
                    if issubclass(type(propulsor_item), RCAIDE.Library.Components.Component):
                        propulsor_item.append_operating_conditions(segment,propulsor)            
            
            for bus_i, bus in enumerate(network.busses):   
                # ------------------------------------------------------------------------------------------------------            
                # Create bus results data structure  
                # ------------------------------------------------------------------------------------------------------
                segment.state.conditions.energy[bus.tag] = RCAIDE.Framework.Mission.Common.Conditions() 
                segment.state.conditions.noise[bus.tag]  = RCAIDE.Framework.Mission.Common.Conditions()   
    
                # ------------------------------------------------------------------------------------------------------
                # Assign network-specific  residuals, unknowns and results data structures
                # ------------------------------------------------------------------------------------------------------
                if bus.active:
                    for propulsor_group in  bus.assigned_propulsors:
                        propulsor =  network.propulsors[propulsor_group[0]]
                        propulsor.append_propulsor_unknowns_and_residuals(segment)
                        
                # ------------------------------------------------------------------------------------------------------
                # Assign sub component results data structures
                # ------------------------------------------------------------------------------------------------------ 
                bus.append_operating_conditions(segment)
                for battery_module in  bus.battery_modules: 
                    battery_module.append_operating_conditions(segment,bus) 
    
                    
                for tag, bus_item in  bus.items():  
                    if issubclass(type(bus_item), RCAIDE.Library.Components.Component):
                        bus_item.append_operating_conditions(segment,bus)                     
    
            for coolant_line_i, coolant_line in enumerate(network.coolant_lines):  
                # ------------------------------------------------------------------------------------------------------            
                # Create coolant_lines results data structure  
                # ------------------------------------------------------------------------------------------------------
                segment.state.conditions.energy[coolant_line.tag] = RCAIDE.Framework.Mission.Common.Conditions()        
                # ------------------------------------------------------------------------------------------------------
                # Assign network-specific  residuals, unknowns and results data structures
                # ------------------------------------------------------------------------------------------------------ 
                 
                for battery_module in coolant_line.battery_modules: 
                    for btms in battery_module:
                        btms.append_operating_conditions(segment,coolant_line)
                        
                for heat_exchanger in coolant_line.heat_exchangers: 
                    heat_exchanger.append_operating_conditions(segment, coolant_line)
                        
                for reservoir in coolant_line.reservoirs: 
                    reservoir.append_operating_conditions(segment, coolant_line)                           

        # Ensure the mission knows how to pack and unpack the unknowns and residuals
        segment.process.iterate.unknowns.network            = self.unpack_unknowns
        segment.process.iterate.residuals.network           = self.residuals        

        return segment
    __call__ = evaluate 


# RCAIDE/Library/Methods/Propulsors/Converters/Shaft_Power_Offtake/compute_shaft_power_offtaker.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------    

# package imports
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
# compute_offtake_shaft_performance
# ----------------------------------------------------------------------------------------------------------------------     
def compute_offtake_shaft_performance(offtake_shaft,offtake_shaft_conditions, conditions):
    """ This computes the work done from the power draw. The following properties are computed: 
    offtake_shaft.outputs.
      power        (numpy.ndarray): power                              [W]
      work_done    (numpy.ndarray): work done normalized by mass flow  [J/(kg/s)] 

    Assumptions:
        None

    Source:
        https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/

    Args:
        offtake_shaft
          .inputs.mdhc                  (numpy.ndarray): Compressor nondimensional mass flow   [unitless] 
          .inputs.reference_temperature (numpy.ndarray): Reference temperature                 [K]
          .inputs.reference_pressure    (numpy.ndarray): Reference pressure                    [Pa]
          .power_draw                           (float): power draw                            [W]

    Returns:
        None 
    """  
    if offtake_shaft.power_draw == 0.0:
        offtake_shaft.outputs.work_done = np.array([0.0]) 
    else: 
        # unpack 
        total_temperature_reference = offtake_shaft_conditions.inputs.total_temperature_reference
        total_pressure_reference    = offtake_shaft_conditions.inputs.total_pressure_reference
        mdhc                        = offtake_shaft_conditions.inputs.mdhc
        Tref                        = offtake_shaft.reference_temperature
        Pref                        = offtake_shaft.reference_pressure
        
        # compute core mass flow rate 
        mdot_core = mdhc * np.sqrt(Tref / total_temperature_reference) * (total_pressure_reference / Pref)

        offtake_shaft_conditions.outputs.power     = offtake_shaft.power_draw
        offtake_shaft_conditions.outputs.work_done = offtake_shaft_conditions.outputs.power / mdot_core  # normalize 
        offtake_shaft_conditions.outputs.work_done[mdot_core == 0] = 0
        
    return 


# RCAIDE/Library/Methods/Propulsors/Converters/supersonic_nozzle/append_supersonic_nozzle_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
# append_supersonic_nozzle_conditions 
# ----------------------------------------------------------------------------------------------------------------------    
def append_supersonic_nozzle_conditions(supersonic_nozzle,segment,propulsor_conditions): 
    propulsor_conditions[supersonic_nozzle.tag]                     = Conditions()
    propulsor_conditions[supersonic_nozzle.tag].inputs              = Conditions()
    propulsor_conditions[supersonic_nozzle.tag].outputs             = Conditions()
    return 


# RCAIDE/Library/Plots/Energy/plot_battery_health_conditions.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style  
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_battery_temperature(results,
                                  save_figure = False,
                                  show_legend = True,
                                  save_filename = "Battery_Temperature",
                                  file_type = ".png",
                                  width = 11, height = 7):
    """Plots the cell-level conditions of the battery throughout flight.

    Assumptions:
    None

    Source:
    None

    Inputs:
    results.segments.conditions.
        freestream.altitude
        weights.total_mass
        weights.vehicle_mass_rate
        frames.body.thrust_force_vector

    Outputs: 
    Plots

    Properties Used:
    N/A	
    """ 
    
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))     

    fig = plt.figure(save_filename)
    fig.set_size_inches(width,height)  
    axis_1 = plt.subplot(2,2,1)
    axis_2 = plt.subplot(2,2,2) 
    axis_3 = plt.subplot(2,2,3)     
 
    for network in results.segments[0].analyses.energy.vehicle.networks: 
        busses  = network.busses 
        for  bus_i, bus in enumerate(busses):
            for b_i, battery in enumerate(bus.battery_modules):
                if b_i == 0 or bus.identical_battery_modules == False:
                    for i in range(len(results.segments)): 
                        bus_results         = results.segments[i].conditions.energy[bus.tag]
                        time                = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min                      
                        battery_conditions  = results.segments[i].conditions.energy[bus.tag].battery_modules[battery.tag]  
                        cell_temp           = battery_conditions.cell.temperature[:,0]
                        cell_charge         = battery_conditions.cell.charge_throughput[:,0]
                        pack_Q              = bus_results.heat_energy_generated[:,0]
                        
                        if b_i == 0 and i == 0:
                            axis_1.plot(time,cell_temp, color = line_colors[i], marker = ps.markers[bus_i], linewidth = ps.line_width, label = battery.tag)
                        else:
                            axis_1.plot(time,cell_temp, color = line_colors[i], marker = ps.markers[bus_i], linewidth = ps.line_width)
                        axis_1.set_ylabel(r'Temperature (K)') 
                        set_axes(axis_1)        
                        
                        axis_2.plot(time, cell_charge, color = line_colors[i], marker = ps.markers[bus_i], linewidth = ps.line_width)
                        axis_2.set_xlabel('Time (mins)')
                        axis_2.set_ylabel(r'Charge Throughput (Ah)')
                        set_axes(axis_2)   
                        
                        axis_3.plot(time, pack_Q/1000, color = line_colors[i], marker = ps.markers[bus_i], linewidth = ps.line_width)
                        axis_3.set_xlabel('Time (mins)')
                        axis_3.set_ylabel(r'$\dot{Q}_{heat}$ (kW)')
                        set_axes(axis_3) 
    
    if show_legend:         
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4) 
    
    # Adjusting the sub-plots for legend
    fig.tight_layout() 
    fig.subplots_adjust(top=0.8)
    
    # set title of plot 
    title_text    = 'Battery Temperature'   
    fig.suptitle(title_text)
    
    if save_figure:
        plt.savefig(save_filename  + file_type)   
    return fig 


# RCAIDE/Library/Components/Propulsors/Turbofan.py 
#
#
# Created:  Mar 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports  
from RCAIDE.Framework.Core      import Container
from .                          import Propulsor
from RCAIDE.Library.Methods.Propulsors.Turbofan_Propulsor.append_turbofan_conditions     import append_turbofan_conditions 
from RCAIDE.Library.Methods.Propulsors.Turbofan_Propulsor.compute_turbofan_performance   import compute_turbofan_performance, reuse_stored_turbofan_data
 
# ---------------------------------------------------------------------------------------------------------------------- 
#  Fan Component
# ---------------------------------------------------------------------------------------------------------------------- 
class Turbofan(Propulsor):
    """
    A turbofan propulsion system model that simulates the performance of a turbofan engine.

    Attributes
    ----------
    tag : str
        Identifier for the turbofan engine. Default is 'Turbofan'.
    
    nacelle : Component
        Nacelle component of the engine. Default is None.
        
    fan : Component
        Fan component of the engine. Default is None.
        
    ram : Component
        Ram inlet component. Default is None.
        
    inlet_nozzle : Component
        Inlet nozzle component. Default is None.
        
    low_pressure_compressor : Component
        Low pressure compressor component. Default is None.
        
    high_pressure_compressor : Component
        High pressure compressor component. Default is None.
        
    low_pressure_turbine : Component
        Low pressure turbine component. Default is None.
        
    high_pressure_turbine : Component
        High pressure turbine component. Default is None.
        
    combustor : Component
        Combustor component. Default is None.
        
    core_nozzle : Component
        Core exhaust nozzle component. Default is None.
        
    fan_nozzle : Component
        Fan exhaust nozzle component. Default is None.
        
    active_fuel_tanks : Container
        Collection of active fuel tanks. Default is None.
        
    engine_diameter : float
        Diameter of the engine [m]. Default is 0.0.
        
    engine_length : float
        Length of the engine [m]. Default is 0.0.
        
    engine_height : float
        Engine centerline height above the ground plane [m]. Default is 0.5.
        
    exa : float
        Distance from fan face to fan exit normalized by fan diameter. Default is 1.0.
        
    plug_diameter : float
        Diameter of the engine plug [m]. Default is 0.1.
        
    geometry_xe : float
        Installation effects geometry parameter. Default is 1.0.
        
    geometry_ye : float
        Installation effects geometry parameter. Default is 1.0.
        
    geometry_Ce : float
        Installation effects geometry parameter. Default is 2.0.
        
    bypass_ratio : float
        Engine bypass ratio. Default is 0.0.
        
    design_isa_deviation : float
        ISA temperature deviation at design point [K]. Default is 0.0.
        
    design_altitude : float
        Design altitude of the engine [m]. Default is 0.0.
        
    SFC_adjustment : float
        Specific fuel consumption adjustment factor (Less than 1 is a reduction). Default is 0.0.
        
    compressor_nondimensional_massflow : float
        Non-dimensional mass flow through the compressor. Default is 0.0.
        
    reference_temperature : float
        Reference temperature for calculations [K]. Default is 288.15.
        
    reference_pressure : float
        Reference pressure for calculations [Pa]. Default is 101325.0.
        
    design_thrust : float
        Design thrust of the engine [N]. Default is 0.0.
        
    mass_flow_rate_design : float
        Design mass flow rate [kg/s]. Default is 0.0.
        
    OpenVSP_flow_through : bool
        Flag for OpenVSP flow-through analysis. Default is False.

    Notes
    -----
    The Turbofan class inherits from the Propulsor base class and implements
    methods for computing turbofan engine performance. It includes functionality
    for handling different operating conditions and performance calculations.

    **Definitions**

    'ISA'
        International Standard Atmosphere - standard atmospheric model

    'SFC'
        Specific Fuel Consumption - fuel efficiency metric

    'OpenVSP'
        Open Vehicle Sketch Pad - open-source parametric aircraft geometry tool

    See Also
    --------
    RCAIDE.Library.Components.Propulsors.Propulsor
    RCAIDE.Library.Components.Propulsors.Turbojet
    RCAIDE.Library.Components.Propulsors.Turboprop
    RCAIDE.Library.Components.Propulsors.Turboshaft
    """
    def __defaults__(self):    
        # setting the default values
        self.tag                                      = 'Turbofan'  
        self.nacelle                                  = None 
        self.fan                                      = None 
        self.ram                                      = None 
        self.inlet_nozzle                             = None 
        self.low_pressure_compressor                  = None 
        self.high_pressure_compressor                 = None 
        self.low_pressure_turbine                     = None 
        self.high_pressure_turbine                    = None 
        self.combustor                                = None 
        self.core_nozzle                              = None 
        self.fan_nozzle                               = None  
        self.active_fuel_tanks                        = None         
        self.engine_diameter                          = 0.0      
        self.engine_length                            = 0.0
        self.engine_height                            = 0.5     # Engine centerline heigh above the ground plane
        self.exa                                      = 1       # distance from fan face to fan exit/ fan diameter)
        self.plug_diameter                            = 0.1     # dimater of the engine plug
        self.geometry_xe                              = 1.      # Geometry information for the installation effects function
        self.geometry_ye                              = 1.      # Geometry information for the installation effects function
        self.geometry_Ce                              = 2.      # Geometry information for the installation effects function
        self.bypass_ratio                             = 0.0 
        self.design_isa_deviation                     = 0.0
        self.design_altitude                          = 0.0
        self.SFC_adjustment                           = 0.0 # Less than 1 is a reduction
        self.compressor_nondimensional_massflow       = 0.0
        self.reference_temperature                    = 288.15
        self.reference_pressure                       = 1.01325*10**5 
        self.design_thrust                            = 0.0
        self.mass_flow_rate_design                    = 0.0
        self.OpenVSP_flow_through                     = False
    
    def append_operating_conditions(self,segment):
        """
        Appends operating conditions to the segment.
        """
        append_turbofan_conditions(self,segment)
        return

    def unpack_propulsor_unknowns(self,segment):   
        return 

    def pack_propulsor_residuals(self,segment): 
        return

    def append_propulsor_unknowns_and_residuals(self,segment): 
        return
    
    def compute_performance(self,state,center_of_gravity = [[0, 0, 0]]):
        """
        Computes turbofan performance including thrust, moment, and power.
        """
        thrust,moment,power,stored_results_flag,stored_propulsor_tag =  compute_turbofan_performance(self,state,center_of_gravity)
        return thrust,moment,power,stored_results_flag,stored_propulsor_tag
    
    def reuse_stored_data(turbofan,state,network,stored_propulsor_tag,center_of_gravity = [[0, 0, 0]]):
        """
        Reuses stored turbofan data for performance calculations.
        """
        thrust,moment,power  = reuse_stored_turbofan_data(turbofan,state,network,stored_propulsor_tag,center_of_gravity)
        return thrust,moment,power 


# RCAIDE/Library/Methods/Weights/Correlations/General_Aviation/__init__.py
# 

"""RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .compute_operating_empty_weight import compute_operating_empty_weight
from .compute_landing_gear_weight    import compute_landing_gear_weight    
from .compute_payload_weight         import compute_payload_weight         
from .compute_systems_weight         import compute_systems_weight         
from .compute_horizontal_tail_weight import compute_horizontal_tail_weight 
from .compute_vertical_tail_weight   import compute_vertical_tail_weight   
from .compute_fuselage_weight        import compute_fuselage_weight        
from .compute_main_wing_weight       import compute_main_wing_weight       



# RCAIDE/Framework/External_Interfaces/OpenVSP/import_vsp_vehicle.py

# Created:  Jun 2018, T. St Francis
# Modified: Aug 2018, T. St Francis
#           Jan 2020, T. MacDonald
#           Jul 2020, E. Botero
#           Sep 2021, R. Erhard
#           Dec 2021, E. Botero

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports 
import RCAIDE
from  RCAIDE.Framework.Core import Units, Data, Container 
from RCAIDE.Framework.External_Interfaces.OpenVSP.vsp_rotor           import read_vsp_rotor
from RCAIDE.Framework.External_Interfaces.OpenVSP.vsp_fuselage         import read_vsp_fuselage
from RCAIDE.Framework.External_Interfaces.OpenVSP.vsp_wing             import read_vsp_wing
from RCAIDE.Framework.External_Interfaces.OpenVSP.vsp_nacelle          import read_vsp_nacelle
from RCAIDE.Framework.External_Interfaces.OpenVSP.get_vsp_measurements import get_vsp_measurements


from copy import deepcopy

try:
    import vsp as vsp
except ImportError:
    try:
        import openvsp as vsp
    except ImportError:
        # This allows RCAIDE to build without OpenVSP
        pass
    
# ---------------------------------------------------------------------------------------------------------------------- 
#  vsp read
# ---------------------------------------------------------------------------------------------------------------------- 
def import_vsp_vehicle(tag,network_type=None, propulsor_type = None, units_type='SI',use_scaling=True,calculate_wetted_area=True): 
    """This reads an OpenVSP vehicle geometry and writes it into a RCAIDE vehicle format.
    Includes wings, fuselages, and rotors.

    Assumptions:
    1. OpenVSP vehicle is composed of conventionally shaped fuselages, wings, and rotors. 
    1a. OpenVSP fuselage: generally narrow at nose and tail, wider in center). 
    1b. Fuselage is designed in VSP as it appears in real life. That is, the VSP model does not rely on
       superficial elements such as canopies, stacks, or additional fuselages to cover up internal lofting oddities.
    1c. This program will NOT account for multiple geometries comprising the fuselage. For example: a wingbox mounted beneath
       is a separate geometry and will NOT be processed.
    2. Fuselage origin is located at nose. VSP file origin can be located anywhere, preferably at the forward tip
       of the vehicle or in front (to make all X-coordinates of vehicle positive). 
    4. Written for OpenVSP 3.21.1

    Source:
    N/A

    Inputs:
    1. A tag for an XML file in format .vsp3.
    2. Units_type set to 'SI' (default) or 'Imperial'
    3. User-specified network
    4. Boolean for whether or not to use the scaling from OpenVSP (default = True).

    Outputs:
    Writes RCAIDE vehicle with these geometries from VSP:    (All values default to SI. Any other 2nd argument outputs Imperial.)
    	Wings.Wing.    (* is all keys)
    		origin                                  [m] in all three dimensions
    		spans.projected                         [m]
    		chords.root                             [m]
    		chords.tip                              [m]
    		aspect_ratio                            [-]
    		sweeps.quarter_chord                    [radians]
    		twists.root                             [radians]
    		twists.tip                              [radians]
    		thickness_to_chord                      [-]
    		dihedral                                [radians]
    		symmetric                               <boolean>
    		tag                                     <string>
    		areas.reference                         [m^2]
    		areas.wetted                            [m^2]
    		Segments.
    		  tag                                   <string>
    		  twist                                 [radians]
    		  percent_span_location                 [-]  .1 is 10%
    		  root_chord_percent                    [-]  .1 is 10%
    		  dihedral_outboard                     [radians]
    		  sweeps.quarter_chord                  [radians]
    		  thickness_to_chord                    [-]
    		  airfoil                               <NACA 4-series, 6 series, or airfoil file>

    	Fuselages.Fuselage.			
    		origin                                  [m] in all three dimensions
    		width                                   [m]
    		lengths.
    		  total                                 [m]
    		  nose                                  [m]
    		  tail                                  [m]
    		heights.
    		  maximum                               [m]
    		  at_quarter_length                     [m]
    		  at_three_quarters_length              [m]
    		effective_diameter                      [m]
    		fineness.nose                           [-] ratio of nose section length to fuselage effective diameter
    		fineness.tail                           [-] ratio of tail section length to fuselage effective diameter
    		areas.wetted                            [m^2]
    		tag                                     <string>
    		segment[].   (segments are in ordered container and callable by number)
    		  vsp.shape                               [point,circle,round_rect,general_fuse,fuse_file]
    		  vsp.xsec_id                             <10 digit string>
    		  percent_x_location
    		  percent_z_location
    		  height
    		  width
    		  length
    		  effective_diameter
    		  tag
    		vsp.xsec_num                              <integer of fuselage segment quantity>
    		vsp.xsec_surf_id                          <10 digit string>

    	Propellers.Propeller.
    		location[X,Y,Z]                            [radians]
    		rotation[X,Y,Z]                            [radians]
    		tip_radius                                 [m]
    	        hub_radius                                 [m] 
 
    """  	

    if isinstance(network_type,RCAIDE.Framework.Networks.Network) != True:
        raise Exception('Vehicle energy network type must be defined. \n Choose from list in RCAIDE.Framework.Networks') 

    if isinstance(propulsor_type,RCAIDE.Library.Components.Propulsors.Propulsor ) != True:
        raise Exception('Vehicle propulsor type must be defined. \n Choose from list in RCAIDE.Library.Compoments.Propulsors')     

    vsp.ClearVSPModel() 
    vsp.ReadVSPFile(tag)	

    vsp_fuselages     = []
    vsp_wings         = []	
    vsp_rotors         = [] 
    vsp_nacelles      = [] 
    vsp_nacelle_type  = []
    
    vsp_geoms         = vsp.FindGeoms()
    geom_names        = []

    vehicle           = RCAIDE.Vehicle()
    vehicle.tag       = tag 

    if units_type == 'SI':
        units_type = 'SI' 
    elif units_type == 'inches':
        units_type = 'inches'	
    else:
        units_type = 'imperial'	

    # The two for-loops below are in anticipation of an OpenVSP API update with a call for GETGEOMTYPE.
    # This print function allows user to enter VSP GeomID manually as first argument in import_vsp_vehicle functions.

    print("VSP geometry IDs: ")	

    # Label each geom type by storing its VSP geom ID. 

    for geom in vsp_geoms: 
        geom_name = vsp.GetGeomName(geom)
        geom_names.append(geom_name)
        print(str(geom_name) + ': ' + geom)
        
    
    # ------------------------------------------------------------------        
    # Use OpenVSP to calculate wetted area
    # ------------------------------------------------------------------
    if calculate_wetted_area:
        measurements = get_vsp_measurements()
        if units_type == 'SI':
            units_factor = Units.meter * 1.
        elif units_type == 'imperial':
            units_factor = Units.foot * 1.
        elif units_type == 'inches':
            units_factor = Units.inch * 1.	         
        
    # ------------------------------------------------------------------
    # AUTOMATIC VSP ENTRY & PROCESSING
    # ------------------------------------------------------------------		

    for geom in vsp_geoms:
        geom_name = vsp.GetGeomName(geom)
        geom_type = vsp.GetGeomTypeName(str(geom))

        if geom_type == 'Fuselage':
            vsp_fuselages.append(geom)
        if geom_type == 'Wing':
            vsp_wings.append(geom)
        if geom_type == 'Propeller':
            vsp_rotors.append(geom) 
        if (geom_type == 'Stack') or (geom_type == 'BodyOfRevolution'):
            vsp_nacelle_type.append(geom_type)
            vsp_nacelles.append(geom) 
        
    # ------------------------------------------------------------------			
    # Read Fuselages 
    # ------------------------------------------------------------------			    
    for fuselage_id in vsp_fuselages:
        sym_planar = vsp.GetParmVal(fuselage_id, 'Sym_Planar_Flag', 'Sym') # Check for symmetry
        sym_origin = vsp.GetParmVal(fuselage_id, 'Sym_Ancestor_Origin_Flag', 'Sym') 
        if sym_planar == 2. and sym_origin == 1.:  
            num_fus  = 2 
            sym_flag = [1,-1]
        else: 
            num_fus  = 1 
            sym_flag = [1] 
        for fux_idx in range(num_fus):	# loop through fuselages on aircraft 
            fuselage = read_vsp_fuselage(fuselage_id,fux_idx,sym_flag[fux_idx],units_type,use_scaling)
            
            if calculate_wetted_area:
                fuselage.areas.wetted = measurements[vsp.GetGeomName(fuselage_id)] * (units_factor**2)
            
            vehicle.append_component(fuselage)
        
    # ------------------------------------------------------------------			    
    # Read Wings 
    # ------------------------------------------------------------------			
    for wing_id in vsp_wings:
        wing = read_vsp_wing(wing_id, units_type,use_scaling)
        if calculate_wetted_area:
            wing.areas.wetted = measurements[vsp.GetGeomName(wing_id)] * (units_factor**2)  
        vehicle.append_component(wing)		 
        
    # ------------------------------------------------------------------			    
    # Read Enegy Network 
    # ------------------------------------------------------------------
    network =  deepcopy(network_type)  
    i       = 0
    # Condition when equal number of rotors and nacelles are defined 
    if len(vsp_rotors) == len(vsp_nacelles):
        for idx , (rotor_id,nacelle_id) in enumerate(zip(vsp_rotors, vsp_nacelles)):
            # define new propulsor 
            propulsor = deepcopy(propulsor_type) 
            propulsor.tag =  'propulsor_' +  str(i+1)
            
            # Rotor 
            rotor           = read_vsp_rotor(rotor_id,units_type)
            rotor.tag       = vsp.GetGeomName(rotor_id) 
            propulsor.rotor = rotor
            
            # Nacelle 
            nacelle = read_vsp_nacelle(nacelle_id,vsp_nacelle_type[idx], units_type)
            if calculate_wetted_area:
                nacelle.areas.wetted = measurements[vsp.GetGeomName(nacelle_id)] * (units_factor**2)           
            propulsor.nacelle = nacelle          
             
            # Append to Network 
            network.propulsors.append(propulsor)
            
            
            # If symmetry is defined
            if vsp.GetParmVal(rotor_id, 'Sym_Planar_Flag', 'Sym')== 2.0:
                # update index 
                i += 1
                
                # define new propulsor 
                propulsor =  deepcopy(propulsor_type) 
                propulsor.tag =  'propulsor_' +  str(i+1)
                
                
                # Rotor 
                rotor_sym = deepcopy(rotor)
                rotor_sym.origin[0][1] = - rotor_sym.origin[0][1] 
                propulsor.rotor = rotor_sym
                
        
                # Nacelle 
                nacelle_sym = deepcopy(nacelle)
                nacelle_sym.origin[0][1] = - nacelle_sym.origin[0][1]  
                nacelle_sym.areas.wetted = nacelle.areas.wetted
                propulsor.nacelle = nacelle          
                 
                # Append to Network             
                network.propulsors.append(propulsor)
                
    # Condition when only nacelles are defined 
    elif (len(vsp_rotors) ==  0) and (len(vsp_nacelles) > 0):

        for idx ,  nacelle_id in enumerate(vsp_nacelles):
            # define new propulsor 
            propulsor = deepcopy(propulsor_type) 
            propulsor.tag =  'propulsor_' +  str(i+1)
             
            # Nacelle 
            nacelle = read_vsp_nacelle(nacelle_id,vsp_nacelle_type[idx], units_type)
            if calculate_wetted_area:
                nacelle.areas.wetted = measurements[vsp.GetGeomName(nacelle_id)] * (units_factor**2)           
            propulsor.nacelle = nacelle          
             
            # Append to Network 
            network.propulsors.append(propulsor)
            
            
            # If symmetry is defined
            if vsp.GetParmVal(nacelle_id, 'Sym_Planar_Flag', 'Sym')== 2.0:
                # update index 
                i += 1
                
                # define new propulsor 
                propulsor =  deepcopy(propulsor_type) 
                propulsor.tag =  'propulsor_' +  str(i+1)
                 
        
                # Nacelle 
                nacelle_sym = deepcopy(nacelle)
                nacelle_sym.origin[0][1] = - nacelle_sym.origin[0][1]  
                nacelle_sym.areas.wetted = nacelle.areas.wetted
                propulsor.nacelle = nacelle          
                 
                # Append to Network             
                network.propulsors.append(propulsor)
                
        

    # Condition when only rotors are defined 
    elif (len(vsp_rotors) >  0) and (len(vsp_nacelles) == 0):
        for idx , rotor_id in enumerate( vsp_rotors):
            # define new propulsor 
            propulsor = deepcopy(propulsor_type) 
            propulsor.tag =  'propulsor_' +  str(i+1)
            
            # Rotor 
            rotor           = read_vsp_rotor(rotor_id,units_type)
            rotor.tag       = vsp.GetGeomName(rotor_id) 
            propulsor.rotor = rotor
            
            # Append to Network 
            network.propulsors.append(propulsor)
            
            
            # If symmetry is defined
            if vsp.GetParmVal(rotor_id, 'Sym_Planar_Flag', 'Sym')== 2.0:
                # update index 
                i += 1
                
                # define new propulsor 
                propulsor =  deepcopy(propulsor_type) 
                propulsor.tag =  'propulsor_' +  str(i+1)
                
                
                # Rotor 
                rotor_sym = deepcopy(rotor)
                rotor_sym.origin[0][1] = - rotor_sym.origin[0][1] 
                propulsor.rotor = rotor_sym  
                propulsor.nacelle = nacelle          
                 
                # Append to Network             
                network.propulsors.append(propulsor) 
            
    else:
        print ('Unequal numbers of rotors and nacelles defined. Skipping propulsor definition.') 
                
    vehicle.networks.append(network)

    return vehicle




# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

import numpy as np

# ----------------------------------------------------------------------
#  Methods
# ----------------------------------------------------------------------

## @ingroup Methods-Geometry-Two_Dimensional-Cross_Section-Planform
def fuselage_planform(fuselage):
    """Calculates fuselage geometry values

    Assumptions:
    None

    Source:
    http://adg.stanford.edu/aa241/drag/wettedarea.html

    Inputs:
    fuselage.
      num_coach_seats       [-]
      seat_pitch            [m]
      seats_abreast         [-]
      fineness.nose         [-]
      fineness.tail         [-]
      lengths.fore_space    [m]
      lengths.aft_space     [m]
      width                 [m]
      heights.maximum       [m]

    Outputs:
    fuselage.
      lengths.nose          [m]
      lengths.tail          [m]
      lengths.cabin         [m]
      lengths.total         [m]
      areas.wetted          [m]
      areas.front_projected [m]
      effective_diameter    [m]

    Properties Used:
    N/A
    """        
    
    # unpack
    number_seats    = fuselage.number_coach_seats
    seat_pitch      = fuselage.seat_pitch
    seats_abreast   = fuselage.seats_abreast
    nose_fineness   = fuselage.fineness.nose
    tail_fineness   = fuselage.fineness.tail
    forward_extra   = fuselage.lengths.fore_space
    aft_extra       = fuselage.lengths.aft_space
    fuselage_width  = fuselage.width
    fuselage_height = fuselage.heights.maximum
    length          = fuselage.lengths.total
    
    if length ==0.:    
        # process
        nose_length     = nose_fineness * fuselage_width
        tail_length     = tail_fineness * fuselage_width
        cabin_length    = number_seats * seat_pitch / seats_abreast + \
                       forward_extra + aft_extra
        fuselage_length = cabin_length + nose_length + tail_length
    else:
        fuselage_length = fuselage.lengths.total
        nose_length     = nose_fineness * fuselage_width
        tail_length     = tail_fineness * fuselage_width      
        cabin_length    = fuselage_length - nose_length - tail_length
        
        if fuselage_length <= 0:
            fuselage_length = 1.
        if nose_length <= 0.:
            nose_length = 1.
        if tail_length <= 0:
            tail_length = 1.
        if cabin_length <= 1.:
            cabin_length = 1.
        
        # Now we can calculate the number of passengers
        number_seats    = np.round(cabin_length * seats_abreast / seat_pitch)
        if number_seats <=0: number_seats=0
    
    wetted_area = 0.0
    
    # model constant fuselage cross section as an ellipse
    # approximate circumference http://en.wikipedia.org/wiki/Ellipse#Circumference
    a = fuselage_width/2.
    b = fuselage_height/2.
    A = np.pi * a * b  # area
    R = (a-b)/(a+b) # effective radius
    C = np.pi*(a+b)*(1.+ ( 3*R**2 )/( 10+np.sqrt(4.-3.*R**2) )) # circumference
    
    wetted_area += C * cabin_length
    cross_section_area = A
    
    # approximate nose and tail wetted area
    # http://adg.stanford.edu/aa241/drag/wettedarea.html
    Deff = (a+b)*(64.-3.*R**4)/(64.-16.*R**2)
    wetted_area += 0.75*np.pi*Deff * (nose_length + tail_length)
    
    # update
    fuselage.lengths.nose          = nose_length
    fuselage.lengths.tail          = tail_length
    fuselage.lengths.cabin         = cabin_length
    fuselage.lengths.total         = fuselage_length
    fuselage.areas.wetted          = wetted_area
    fuselage.areas.front_projected = cross_section_area
    fuselage.effective_diameter    = Deff
    fuselage.number_coach_seats    = number_seats
    
    return fuselage



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Raymer/compute_landing_gear_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import RCAIDE
from RCAIDE.Framework.Core    import Units ,  Data 
 
# ----------------------------------------------------------------------------------------------------------------------
#  Horizontal Tail Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_landing_gear_weight(vehicle):
    """ Calculate the weight of the landing gear of a transport aircraft based on the Raymer method

        Assumptions:
            No fuselage mounted landing gear
            1 cargo door
            gear load factor = 3
            number of main gear shock struts = 2
            stall speed = 51 kts as defined by FAA
            Not a reciprocating engine

        Source:
            Aircraft Design: A Conceptual Approach (2nd edition)

        Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.design_range: design range of aircraft                        [m]
                -.mass_properties.max_takeoff: MTOW                             [kg]
                -.systems.accessories: type of aircraft (short-range, commuter
                                                        medium-range, long-range,
                                                        sst, cargo)
                -.landing_gear.main_strut_length: main strut length              [m]
                -.landing_gear.main_wheels: number of wheels on main landing gear
                -.landing_gear.nose_strut_length: nose strut length              [m]
                -.landing_gear.nose_wheels: number of wheels on nose landing gear

            fuse - data dictionary with specific fuselage properties            [dimensionless]

        Outputs:
            weight_fuse - weight of the fuselage                                [kilograms]

        Properties Used:
            N/A
    """
    Kmp         = 1  # assuming not a kneeling gear
    if vehicle.systems.accessories == "sst":
        RFACT   = 0.00009
    else:
        RFACT   = 0.00004
    DESRNG      = vehicle.flight_envelope.design_range / Units.nmi  # Design range in nautical miles
    WLDG        = vehicle.mass_properties.max_takeoff / Units.lbs * (1 - RFACT * DESRNG)
    Ngear       = 3  # gear load factor, usually around 3
    Nl          = Ngear * 1.5  # ultimate landing load factor
      
    Nmw  = 0
    Lm   = 0 
    Ln   = 0
    Nnw  = 0
    Nmss = 2  # number of main gear shock struts assumed to be 2
    for landing_gear in  vehicle.landing_gears:
        if isinstance(landing_gear,RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear):
            Nmw  = landing_gear.wheels * Nmss    
            Lm   = landing_gear.strut_length / Units.inch 
        elif isinstance(landing_gear,RCAIDE.Library.Components.Landing_Gear.Nose_Landing_Gear): 
            Ln          = landing_gear.strut_length / Units.inch
            Nnw         = landing_gear.wheels  
        
    Vstall      = 51 * Units.kts  # stall speed
    Knp         = 1  # assuming not a reciprocating engine
    W_main_landing_gear = 0.0106 * Kmp * WLDG ** 0.888 * Nl ** 0.25 * Lm ** 0.4 * Nmw ** 0.321 * Nmss ** (-0.5) * Vstall ** 0.1
    W_nose_landing_gear = 0.032 * Knp * WLDG ** 0.646 * Nl ** 0.2 * Ln ** 0.5 * Nnw ** 0.45

    output          = Data()
    output.main     = W_main_landing_gear * Units.lbs
    output.nose     = W_nose_landing_gear * Units.lbs
    return output



# RCAIDE/Methods/Energy/Propulsors/Turboshaft_Propulsor/size_core.py
# 
# 
# Created:  Jul 2023, M. Clarke
# Modified: Jun 2024, M. Guidotti & D.J. Lee

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from RCAIDE.Library.Methods.Propulsors.Turboshaft_Propulsor import compute_power

# Python package imports
import numpy                                                       as np

# ----------------------------------------------------------------------------------------------------------------------
#  size_core
# ----------------------------------------------------------------------------------------------------------------------
def size_core(turboshaft,turboshaft_conditions,conditions):
    """Sizes the core flow for the design condition.

    Assumptions:
    Perfect gas
    Turboshaft engine with free power turbine

    Sources:
    [1] https://soaneemrana.org/onewebmedia/ELEMENTS%20OF%20GAS%20TURBINE%20PROPULTION2.pdf - Page 332 - 336
    [2] https://www.colorado.edu/faculty/kantha/sites/default/files/attached-files/70652-116619_-_luke_stuyvenberg_-_dec_17_2015_1258_pm_-_stuyvenberg_helicopterturboshafts.pdf

    Inputs:
    conditions.freestream.speed_of_sound [m/s] (conditions is also passed to turboshaft.compute(..))
    turboshaft.inputs.
      bypass_ratio                            [-]
      total_temperature_reference             [K]
      total_pressure_reference                [Pa]
      number_of_engines                       [-]

    Outputs:
    turboshaft.outputs.non_dimensional_power  [-]

    Properties Used:
    turboshaft.
      reference_temperature                   [K]
      reference_pressure                      [Pa]
      total_design                            [W] - Design power
    """             
    
    #unpack from turboshaft
    Tref                                           = turboshaft.reference_temperature
    Pref                                           = turboshaft.reference_pressure 
    total_temperature_reference                    = turboshaft_conditions.total_temperature_reference  
    total_pressure_reference                       = turboshaft_conditions.total_pressure_reference 

    #compute nondimensional power
    compute_power(turboshaft,turboshaft_conditions,conditions)

    #unpack results 
    Psp                                            = turboshaft_conditions.non_dimensional_power
    
    #compute dimensional mass flow rates
    mdot_air                                       = turboshaft.design_power/Psp
    mdot_compressor                                = mdot_air/ (np.sqrt(Tref/total_temperature_reference)*(total_pressure_reference/Pref))

    #pack outputs
    turboshaft.mass_flow_rate_design               = mdot_air
    turboshaft.compressor.mass_flow_rate           = mdot_compressor

    return    



## @ingroup Attributes-Solids

# Aluminum_Rib.py
#
# Created: Jul, 2017, J. Smart
# Modified: Apr, 2018, J. Smart

#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------

from .Solid import Solid
from .Aluminum import Aluminum
from RCAIDE.Framework.Core import Units

#-------------------------------------------------------------------------------
# Aluminim Component Material Property Data Class
#-------------------------------------------------------------------------------

## @ingroup Attributes-Solids
class Aluminum_Rib(Aluminum):
    """ Physical Constants Specific to 6061-T6 Aluminum Ribs
    
    Assumptions:
    Limit of machining capability for precision components
    
    Source:
    None
    
    Inputs:
    N/A
    
    Outputs:
    N/A
    
    Properties Used:
    None
    """

    def __defaults__(self):
        """Sets material properties at instantiation.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        N/A

        Outputs:
        N/A

        Properties Used:
        None
        """


        self.minimum_gage_thickness = 1.5e-3   * Units.m
        self.minimum_width          = 25.4e-3  * Units.m




# RCAIDE/Methods/Energy/Propulsors/Networks/turboshaft_Propulsor/compute_turboshaft_performance.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports      
from RCAIDE.Framework.Core import Data   
from RCAIDE.Library.Methods.Propulsors.Converters.Ram                import compute_ram_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Combustor          import compute_combustor_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Compressor         import compute_compressor_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Turbine            import compute_turbine_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Expansion_Nozzle   import compute_expansion_nozzle_performance 
from RCAIDE.Library.Methods.Propulsors.Converters.Compression_Nozzle import compute_compression_nozzle_performance
from RCAIDE.Library.Methods.Propulsors.Turboshaft_Propulsor          import compute_power
 
# python imports 
from copy import deepcopy 
# ----------------------------------------------------------------------------------------------------------------------
# compute_turboshaft_performance
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_turboshaft_performance(turboshaft,state,center_of_gravity= [[0.0, 0.0,0.0]]):    
    ''' Computes the perfomrance of one turboshaft
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    conditions           - operating conditions data structure     [-]  
    fuel_line            - fuelline                                [-] 
    turboshaft           - turboshaft data structure               [-] 
    total_power          - power of turboshaft group               [W] 

    Outputs:  
    total_power          - power of turboshaft group               [W] 
    stored_results_flag  - boolean for stored results              [-]     
    stored_propulsor_tag - name of turboshaft with stored results  [-]
    
    Properties Used: 
    N.A.        
    ''' 
    conditions                = state.conditions 
    noise_conditions          = conditions.noise[turboshaft.tag]  
    turboshaft_conditions     = conditions.energy[turboshaft.tag]
    rho                       = conditions.freestream.density  
    ram                       = turboshaft.ram
    inlet_nozzle              = turboshaft.inlet_nozzle
    compressor                = turboshaft.compressor
    combustor                 = turboshaft.combustor
    high_pressure_turbine     = turboshaft.high_pressure_turbine
    low_pressure_turbine      = turboshaft.low_pressure_turbine 
    core_nozzle               = turboshaft.core_nozzle

    # unpack component conditions 
    ram_conditions          = turboshaft_conditions[ram.tag]     
    inlet_nozzle_conditions = turboshaft_conditions[inlet_nozzle.tag]
    core_nozzle_conditions  = turboshaft_conditions[core_nozzle.tag] 
    compressor_conditions   = turboshaft_conditions[compressor.tag] 
    lpt_conditions          = turboshaft_conditions[low_pressure_turbine.tag]
    hpt_conditions          = turboshaft_conditions[high_pressure_turbine.tag]
    combustor_conditions    = turboshaft_conditions[combustor.tag] 

    # Step 1: Set the working fluid to determine the fluid properties
    ram.working_fluid                             = turboshaft.working_fluid

    # Step 2: Compute flow through the ram , this computes the necessary flow quantities and stores it into conditions
    compute_ram_performance(ram,ram_conditions,conditions)

    # Step 3: link inlet nozzle to ram 
    inlet_nozzle_conditions.inputs.stagnation_temperature             = ram_conditions.outputs.stagnation_temperature
    inlet_nozzle_conditions.inputs.stagnation_pressure                = ram_conditions.outputs.stagnation_pressure
    inlet_nozzle_conditions.inputs.static_temperature                 = ram_conditions.outputs.static_temperature
    inlet_nozzle_conditions.inputs.static_pressure                    = ram_conditions.outputs.static_pressure
    inlet_nozzle_conditions.inputs.mach_number                        = ram_conditions.outputs.mach_number
    inlet_nozzle.working_fluid                                        = ram.working_fluid

    # Step 4: Compute flow through the inlet nozzle
    compute_compression_nozzle_performance(inlet_nozzle,inlet_nozzle_conditions,conditions)      

    # Step 5: Link low pressure compressor to the inlet nozzle 
    compressor_conditions.inputs.stagnation_temperature  = inlet_nozzle_conditions.outputs.stagnation_temperature
    compressor_conditions.inputs.stagnation_pressure     = inlet_nozzle_conditions.outputs.stagnation_pressure
    compressor_conditions.inputs.static_temperature      = inlet_nozzle_conditions.outputs.static_temperature
    compressor_conditions.inputs.static_pressure         = inlet_nozzle_conditions.outputs.static_pressure
    compressor_conditions.inputs.mach_number             = inlet_nozzle_conditions.outputs.mach_number  
    compressor.working_fluid                             = inlet_nozzle.working_fluid 

    # Step 6: Compute flow through the low pressure compressor
    compute_compressor_performance(compressor,compressor_conditions,conditions)

    # Step 11: Link the combustor to the high pressure compressor
    combustor_conditions.inputs.stagnation_temperature                = compressor_conditions.outputs.stagnation_temperature
    combustor_conditions.inputs.stagnation_pressure                   = compressor_conditions.outputs.stagnation_pressure
    combustor_conditions.inputs.static_temperature                    = compressor_conditions.outputs.static_temperature
    combustor_conditions.inputs.static_pressure                       = compressor_conditions.outputs.static_pressure
    combustor_conditions.inputs.mach_number                           = compressor_conditions.outputs.mach_number  
    combustor.working_fluid                                           = compressor.working_fluid 

    # Step 12: Compute flow through the high pressor compressor 
    compute_combustor_performance(combustor,combustor_conditions,conditions)

    #link the high pressure turbione to the combustor 
    hpt_conditions.inputs.stagnation_temperature    = combustor_conditions.outputs.stagnation_temperature
    hpt_conditions.inputs.stagnation_pressure       = combustor_conditions.outputs.stagnation_pressure
    hpt_conditions.inputs.fuel_to_air_ratio         = combustor_conditions.outputs.fuel_to_air_ratio 
    hpt_conditions.inputs.static_temperature        = combustor_conditions.outputs.static_temperature
    hpt_conditions.inputs.static_pressure           = combustor_conditions.outputs.static_pressure
    hpt_conditions.inputs.mach_number               = combustor_conditions.outputs.mach_number 
    hpt_conditions.inputs.compressor                = compressor_conditions.outputs 
    hpt_conditions.inputs.bypass_ratio              = 0.0
    hpt_conditions.inputs.fan                       = Data()
    hpt_conditions.inputs.fan.work_done             = 0.0
    high_pressure_turbine.working_fluid             = combustor.working_fluid
    
    compute_turbine_performance(high_pressure_turbine,hpt_conditions,conditions)

    #link the low pressure turbine to the high pressure turbine 
    lpt_conditions.inputs.stagnation_temperature     = hpt_conditions.outputs.stagnation_temperature
    lpt_conditions.inputs.stagnation_pressure        = hpt_conditions.outputs.stagnation_pressure 
    lpt_conditions.inputs.static_temperature         = hpt_conditions.outputs.static_temperature
    lpt_conditions.inputs.static_pressure            = hpt_conditions.outputs.static_pressure 
    lpt_conditions.inputs.mach_number                = hpt_conditions.outputs.mach_number  
    low_pressure_turbine.working_fluid               = high_pressure_turbine.working_fluid    
    lpt_conditions.inputs.compressor                 = Data()
    lpt_conditions.inputs.compressor.work_done       = 0.0     
    lpt_conditions.inputs.fuel_to_air_ratio          = combustor_conditions.outputs.fuel_to_air_ratio 
    lpt_conditions.inputs.bypass_ratio               = 0.0
    lpt_conditions.inputs.fan                        = Data()
    lpt_conditions.inputs.fan.work_done              = 0.0

    compute_turbine_performance(low_pressure_turbine,lpt_conditions,conditions)

    #link the core nozzle to the low pressure turbine
    core_nozzle_conditions.inputs.stagnation_temperature     = lpt_conditions.outputs.stagnation_temperature
    core_nozzle_conditions.inputs.stagnation_pressure        = lpt_conditions.outputs.stagnation_pressure
    core_nozzle_conditions.inputs.static_temperature         = lpt_conditions.outputs.static_temperature
    core_nozzle_conditions.inputs.static_pressure            = lpt_conditions.outputs.static_pressure  
    core_nozzle_conditions.inputs.mach_number                = lpt_conditions.outputs.mach_number   
    core_nozzle.working_fluid                                = low_pressure_turbine.working_fluid  

    #flow through the core nozzle
    compute_expansion_nozzle_performance(core_nozzle,core_nozzle_conditions,conditions) 
 
    # Link the thrust component to the core nozzle
    turboshaft_conditions.core_exit_velocity                       = core_nozzle_conditions.outputs.velocity
    turboshaft_conditions.core_area_ratio                          = core_nozzle_conditions.outputs.area_ratio
    turboshaft_conditions.core_nozzle                              = core_nozzle_conditions.outputs

    # Link the thrust component to the combustor
    turboshaft_conditions.fuel_to_air_ratio                        = combustor_conditions.outputs.fuel_to_air_ratio 

    # Link the thrust component to the low pressure compressor 
    turboshaft_conditions.combustor_stagnation_temperature         = combustor_conditions.outputs.stagnation_temperature
    turboshaft_conditions.total_temperature_reference              = compressor_conditions.inputs.stagnation_temperature
    turboshaft_conditions.total_pressure_reference                 = compressor_conditions.inputs.stagnation_pressure 
    turboshaft_conditions.flow_through_core                        =  1.0 #scaled constant to turn on core thrust computation
    turboshaft_conditions.flow_through_fan                         =  0.0 #scaled constant to turn on fan thrust computation        
    
    # Compute the power
    compute_power(turboshaft,turboshaft_conditions,conditions) 

    # Store data
    core_nozzle_res = Data(
                exit_static_temperature             = core_nozzle_conditions.outputs.static_temperature,
                exit_static_pressure                = core_nozzle_conditions.outputs.static_pressure,
                exit_stagnation_temperature         = core_nozzle_conditions.outputs.stagnation_temperature,
                exit_stagnation_pressure            = core_nozzle_conditions.outputs.static_pressure,
                exit_velocity                       = core_nozzle_conditions.outputs.velocity
            )
  
    noise_conditions.turboshaft.core_nozzle   = core_nozzle_res  
    
    # Pack results   
    moment                 = 0*state.ones_row(3)
    thrust                 = 0*state.ones_row(3) 
    power                  = turboshaft_conditions.power  
    stored_results_flag    = True
    stored_propulsor_tag   = turboshaft.tag
    
    return thrust,moment,power,stored_results_flag,stored_propulsor_tag

def reuse_stored_turboshaft_data(turboshaft,state,network,stored_propulsor_tag,center_of_gravity= [[0.0, 0.0,0.0]]):
    '''Reuses results from one turboshaft for identical propulsors
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    conditions           - operating conditions data structure     [-]  
    fuel_line            - fuelline                                [-] 
    turboshaft            - turboshaft data structure              [-] 
    total_power          - power of turboshaft group               [W] 

    Outputs:  
    total_power          - power of turboshaft group               [W] 
    
    Properties Used: 
    N.A.        
    ''' 
    conditions                         = state.conditions   
    conditions.energy[turboshaft.tag]  = deepcopy(conditions.energy[stored_propulsor_tag])
    conditions.noise[turboshaft.tag]   = deepcopy(conditions.noise[stored_propulsor_tag])
      
    power    = conditions.energy[turboshaft.tag].power    
    moment   = 0*state.ones_row(3)
    thrust   = 0*state.ones_row(3)
    return thrust,moment,power    


# RCAIDE/Library/Methods/Thermal_Management/Heat_Exchangers/Cross_Flow_Heat_Exchanger/compute_heat_exhanger_factors.py
# 
# Created : Dec 2022, C.R. Zhao
# Modified: Jun 2023, M. Clarke

import numpy as np
from scipy import interpolate

# ----------------------------------------------------------------------
#  Methods
# ---------------------------------------------------------------------- 
def compute_heat_exhanger_factors(kc_vals,ke_vals,delta, Re):
    '''
    Core library about Kc & Ke factors for different types of HEX"
    
     Inputs:  
             kc_vals
             ke_vals
             Re
              
          Outputs:   
             Kc
             Ke
              
          Assumptions: 
             N/A 
        
          Source:
             None
    
    '''
    Kc = np.zeros(shape=Re.shape)
    for i in range(Kc.shape[0]):
        if Re[i] < 2000:
            Kc[i] = -0.4252 * delta * delta + 0.0326 * delta + 1.1798
        elif Re[i] > 1e4:
            Kc[i] = -0.419 * delta * delta + 0.0208 * delta + 0.4033
        else:
            delta_set = kc_vals[:, 0]
            Re_set = [2e3, 3e3, 5e3, 1e4]
            Kc_set_T = np.array([(kc_vals[:, 2]), (kc_vals[:, 3]), (kc_vals[:, 4]), (kc_vals[:, 5])])
            f = interpolate.interp2d(delta_set, Re_set, Kc_set_T, kind='linear')
            Kc[i] = f(delta, Re[i])
    
    Ke = np.zeros(shape=Re.shape)
    for i in range(Ke.shape[0]):
        if Re[i] < 2000:
            Ke[i] = 0.996 * delta * delta - 2.7687 * delta + 1.0016
        elif Re[i] > 1e4:
            Ke[i] = 0.9832 * delta * delta - 1.9823 * delta + 1
        else:
            delta_set = ke_vals[:, 0]
            Re_set = [2e3, 3e3, 5e3, 1e4]
            Ke_set_T = np.array([(ke_vals[:, 2]), (ke_vals[:, 3]), (ke_vals[:, 4]), (ke_vals[:, 5])])
            f = interpolate.interp2d(delta_set, Re_set, Ke_set_T, kind='linear')
            Ke[i] = f(delta, Re[i])
    return Kc, Ke


# RCAIDE/Library/Methods/Aerodynamics/Common/Drag/lift_wave_drag.py 
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports    
from .wave_drag  import wave_drag  

# package imports
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  Compressibility Drag Total
# ---------------------------------------------------------------------------------------------------------------------- 
def lift_wave_drag(conditions,wing,Sref):
    """Determine lift wave drag for supersonic speeds.

    Assumptions: 

    Source:
        Stanford AA241 A/B Course Notes

    Args:
        conditions.freestream.mach_number (numpy.array): mach number              [unitless] 
        wing.areas.reference                    (float): wing reference area      [m^2]
        Sref                                    (float): main wing reference area [m^2]

    Returns:
        cd_c_l                            (numpy.array): Wave drag CD due to lift [unitless]
        """
    
    cd_lift_wave         = wave_drag(conditions,wing) 
    mach                 = conditions.freestream.mach_number 
    cd_c_l               = np.zeros_like(mach)  
    cd_c_l[mach >= 1.01] = cd_lift_wave[0:len(mach[mach >= 1.01]),0] 
    cd_c_l               = cd_c_l*wing.areas.reference/Sref

    return cd_c_l


# RCAIDE/Library/Plots/Performance/Aerodynamics/plot_aerodynamic_coefficients.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_aerodynamic_coefficients(results,
                             save_figure = False,  
                             show_legend = True,
                             save_filename = "Aerodynamic_Coefficents",
                             file_type = ".png",
                             width = 11, height = 7):
    """This plots the aerodynamic coefficients
    
    Assumptions:
    None
    
    Source:
    None
    
    Inputs:
    results.segments.condtions.aerodynamics.
        lift_coefficient
        drag_coefficient
        angle_of_attack
        
    Outputs:
    Plots
    
    Properties Used:
    N/A
    """ 

    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))     
     
    fig   = plt.figure(save_filename)
    fig.set_size_inches(width,height)
     
    for i in range(len(results.segments)): 
        time = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min
        cl   = results.segments[i].conditions.aerodynamics.coefficients.lift.total[:,0,None]
        cd   = results.segments[i].conditions.aerodynamics.coefficients.drag.total[:,0,None]
        aoa  = results.segments[i].conditions.aerodynamics.angles.alpha[:,0] / Units.deg
        l_d  = cl/cd    
                       
        segment_tag  =  results.segments[i].tag
        segment_name = segment_tag.replace('_', ' ')
        axis_1 = plt.subplot(2,2,1) 
        axis_1.plot(time, aoa, color = line_colors[i], marker = ps.markers[0],markersize = ps.marker_size, linewidth = ps.line_width, label = segment_name) 
        axis_1.set_ylabel(r'AoA (deg)') 
        axis_1.set_xlabel('Time (mins)')        
        axis_1.set_ylim([-5,15])
        set_axes(axis_1)    

        axis_2 = plt.subplot(2,2,2)        
        axis_2.plot(time, l_d, color = line_colors[i], marker = ps.markers[0],markersize = ps.marker_size, linewidth = ps.line_width) 
        axis_2.set_ylabel(r'L/D')
        axis_2.set_xlabel('Time (mins)')
        set_axes(axis_2) 

        axis_3 = plt.subplot(2,2,3) 
        axis_3.plot(time, cl, color = line_colors[i], marker = ps.markers[0],markersize = ps.marker_size, linewidth = ps.line_width)
        axis_3.set_xlabel('Time (mins)')
        axis_3.set_ylabel(r'$C_L$')
        set_axes(axis_3) 

        axis_4 = plt.subplot(2,2,4)        
        axis_4.plot(time, cd, color = line_colors[i], marker = ps.markers[0],markersize = ps.marker_size, linewidth = ps.line_width)
        axis_4.set_xlabel('Time (mins)')
        axis_4.set_ylabel(r'$C_D$')
        axis_4.set_ylim([0,0.1])
        set_axes(axis_4) 
        
    if show_legend:
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4)  
    
    # Adjusting the sub-plots for legend
    fig.tight_layout() 
    fig.subplots_adjust(top=0.8)  
    
    if save_figure:
        fig.savefig(save_filename   + file_type)  
    return fig 


# RCAIDE/Library/Missions/Common/Update/orientations.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
from RCAIDE.Framework.Core  import  angles_to_dcms, orientation_product, orientation_transpose

# package imports 
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Update Orientations
# ----------------------------------------------------------------------------------------------------------------------
def orientations(segment):
    
    """ Updates the orientation of the vehicle throughout the mission for each relevant axis
    
        Assumptions:
        This assumes the vehicle has 3 frames: inertial, body, and wind 
        
        Inputs:
        segment.state.conditions:
            frames.inertial.velocity_vector          [meters/second]
            frames.body.inertial_rotations           [Radians]
        segment.analyses.planet.mean_radius [meters]
        state.numerics.time.integrate                [float]
            
        Outputs:
            segment.state.conditions:           
                aerodynamics.angles.alpha         [Radians]
                aerodynamics.angles.beta          [Radians]
                aerodynamics.angles.roll          [Radians]
                frames.body.transform_to_inertial [Radians]
                frames.wind.body_rotations        [Radians]
                frames.wind.transform_to_inertial [Radians]
    

        Properties Used:
        N/A
    """

    # unpack
    conditions              = segment.state.conditions
    V_inertial              = conditions.frames.inertial.velocity_vector
    body_inertial_rotations = conditions.frames.body.inertial_rotations 
    roll_rate               = segment.state.conditions.static_stability.roll_rate      
    pitch_rate              = segment.state.conditions.static_stability.pitch_rate 
    yaw_rate                = segment.state.conditions.static_stability.yaw_rate    
    
    # ------------------------------------------------------------------
    #  Body Frame
    # ------------------------------------------------------------------

    # body frame rotations
    phi = body_inertial_rotations[:,0,None] 

    # body frame tranformation matrices
    T_inertial2body = angles_to_dcms(body_inertial_rotations,(2,1,0))
    T_body2inertial = orientation_transpose(T_inertial2body)

    # transform inertial velocity to body frame
    V_body = orientation_product(T_inertial2body,V_inertial)

    # project inertial velocity into body x-z plane
    V_stability = V_body * 1.

    # calculate angle of attack
    alpha = np.arctan2(V_stability[:,2],V_stability[:,0])[:,None]

    # calculate side slip
    beta = np.arctan2(V_stability[:,1],V_stability[:,0])[:,None]

    # pack aerodynamics angles
    conditions.aerodynamics.angles.alpha[:,0] = alpha[:,0]
    conditions.aerodynamics.angles.beta[:,0]  = beta[:,0]
    conditions.aerodynamics.angles.phi[:,0]   = phi[:,0]

    # pack transformation tensor
    conditions.frames.body.transform_to_inertial = T_body2inertial 

    # ------------------------------------------------------------------
    #  Wind Frame
    # ------------------------------------------------------------------

    # back calculate wind frame rotations
    wind_body_rotations = body_inertial_rotations * 0.
    wind_body_rotations[:,0] = 0          # no roll in wind frame
    wind_body_rotations[:,1] = alpha[:,0] # theta is angle of attack
    wind_body_rotations[:,2] = beta[:,0]  # beta is side slip angle

    # wind frame tranformation matricies
    T_wind2body     = angles_to_dcms(wind_body_rotations,(2,1,0)) 
    T_wind2inertial = orientation_product(T_wind2body,T_body2inertial)

    # pack wind rotations
    conditions.frames.wind.body_rotations = wind_body_rotations

    # pack transformation tensor
    conditions.frames.wind.transform_to_inertial = T_wind2inertial
    
    # ------------------------------------------------------------------
    # Rotation rates 
    # ------------------------------------------------------------------ 
    stability_frame_rotations       =  np.concatenate((np.concatenate((roll_rate, pitch_rate), axis=1), yaw_rate), axis=1)
    phi                             = body_inertial_rotations[:, 0]
    theta                           = body_inertial_rotations[:, 1]
    reverse_transformation          = np.zeros_like(T_body2inertial)
    reverse_transformation[:, 0, 0] =  1 
    reverse_transformation[:, 0, 1] =  np.sin(phi)*np.tan(theta)
    reverse_transformation[:, 0, 2] =  np.cos(phi)*np.tan(theta)
    reverse_transformation[:, 1, 1] =  np.cos(phi)
    reverse_transformation[:, 1, 2] =  -np.sin(phi) 
    reverse_transformation[:, 2, 1] =  np.sin(phi) /np.cos(theta)
    reverse_transformation[:, 2, 2] =  np.cos(phi) /np.cos(theta) 
    inertial_rotations              =  orientation_product(reverse_transformation,stability_frame_rotations)
    segment.state.conditions.frames.inertial.angular_velocity_vector = inertial_rotations 
    
    return
         



# ----------------------------------------------------------------------
#  Compute Span Location from Chord Length
# ---------------------------------------------------------------------- 

## @ingroup Methods-Geometry-Three_Dimensional
def compute_span_location_from_chord_length(wing,chord_length):
    """Computes the location along the half-span given a chord length.

    Assumptions:
    Linear variation of chord with span. Returns 0 if constant chord wing.

    Source:
    None

    Inputs:
    wing.chords.
      root                [m]
      tip                 [m]
    wing.spans.projected  [m]
    chord_length          [m]

    Outputs:
    span_location         [m] 

    Properties Used:
    N/A
    """      
    
    #unpack
    ct = wing.chords.tip
    cr = wing.chords.root
    b  = wing.spans.projected
    
    b_2 = b/2.
    
    if (cr-ct)==0:
        span_location = 0.
    else:
        span_location = b_2*(1-(chord_length-ct)/(cr-ct))
    
        
    return span_location


# RCAIDE/Library/Methods/Aerodynamics/Athena_Vortex_Lattice/__init__.py
""" RCAIDE AVL Interface Package Setup
"""

from .evaluate_AVL              import  *
from .build_AVL_surrogates      import build_AVL_surrogates
from .train_AVL_surrogates      import train_AVL_surrogates 
from .create_avl_datastructures import translate_avl_wing, translate_avl_body , populate_wing_sections, populate_body_sections
from .purge_files               import purge_files
from .read_results              import read_results
from .run_AVL_analysis          import run_AVL_analysis
from .translate_data            import translate_conditions_to_cases, translate_results_to_conditions
from .write_geometry            import write_geometry
from .write_mass_file           import write_mass_file
from .write_input_deck          import write_input_deck
from .write_run_cases           import write_run_cases
from .write_avl_airfoil_file    import write_avl_airfoil_file 

from . import AVL_Objects



# RCAIDE/External_Interfaces/OpenVSP/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from .export_vsp_vehicle       import * 
from .get_fuel_tank_properties import *
from .get_vsp_measurements     import *
from .import_vsp_vehicle       import * 
from .mach_slices              import mach_slices
from .vsp_boom                 import *
from .vsp_fuselage             import *
from .vsp_nacelle              import *
from .vsp_wing                 import *
from .vsp_rotor                import *
from .write_vsp_mesh           import write_vsp_mesh


# RCAIDE/Library/Missions/Segments/Climb/Constant_Speed_Constant_Angle.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# Package imports  
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.
    
    Assumptions:
    Constant true airspeed, with a constant climb angle

    Source:
    N/A

    Inputs:
    segment.climb_angle                         [radians]
    segment.air_speed                           [meter/second]
    segment.altitude_start                      [meters]
    segment.altitude_end                        [meters]
    state.numerics.dimensionless.control_points [Unitless]
    conditions.freestream.density               [kilograms/meter^3]

    Outputs:
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]

    Properties Used:
    N/A
    """        
    
    # unpack
    climb_angle = segment.climb_angle
    air_speed   = segment.air_speed   
    alt0        = segment.altitude_start 
    altf        = segment.altitude_end
    beta        = segment.sideslip_angle
    t_nondim    = segment.state.numerics.dimensionless.control_points
    conditions  = segment.state.conditions  

    # check for initial velocity
    if air_speed is None: 
        if not segment.state.initials: raise AttributeError('airspeed not set')
        air_speed = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])
        
    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]

    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0
    
    # process velocity vector
    v_mag = air_speed
    v_x   = np.cos(beta)*v_mag * np.cos(climb_angle)
    v_y   = np.sin(beta)*v_mag * np.cos(climb_angle)
    v_z   = -v_mag * np.sin(climb_angle)
    
    # pack conditions    
    conditions.frames.inertial.velocity_vector[:,0] = v_x
    conditions.frames.inertial.velocity_vector[:,1] = v_y
    conditions.frames.inertial.velocity_vector[:,2] = v_z
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0] # z points down
    conditions.freestream.altitude[:,0]             =  alt[:,0] # positive altitude in this context


# RCAIDE/Library/Plots/Geometry/plot_rotor.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units

from plotly.subplots import make_subplots
import pandas as pd
import plotly.graph_objects as go

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------  
def plot_rotor(prop, face_color = 'red', edge_color = 'black' ,show_figure = True, save_figure = False, save_filename = "Propeller_Geometry", file_type = ".png"):
    """This plots the geometry of a propeller or rotor

    Assumptions:
    None

    Source:
    None

    Inputs:
    RCAIDE.Library.Components.Energy.Converters.Propeller()

    Outputs:
    Plots

    Properties Used:
    N/A
    """
    # Initialize figure
    fig = make_subplots(rows=2, cols=2)

    df1 = pd.DataFrame(dict(x=prop.radius_distribution, y=prop.twist_distribution/Units.degrees))
    df2 = pd.DataFrame(dict(x=prop.radius_distribution, y=prop.chord_distribution))
    df3 = pd.DataFrame(dict(x=prop.radius_distribution, y=prop.max_thickness_distribution))
    df4 = pd.DataFrame(dict(x=prop.radius_distribution, y=prop.mid_chord_alignment))

    fig.append_trace(go.Line(df1), row=1, col=1)
    fig.append_trace(go.Line(df2), row=1, col=2)
    fig.append_trace(go.Line(df3), row=2, col=1)
    fig.append_trace(go.Line(df4), row=2, col=2)

    fig.update_xaxes(title_text="Radial Station", row=1, col=1)
    fig.update_yaxes(title_text="Twist (Deg)", row=1, col=1)
    fig.update_xaxes(title_text="Radial Station", row=1, col=2)
    fig.update_yaxes(title_text="Chord (m)", row=1, col=2)
    fig.update_xaxes(title_text="Radial Station", row=2, col=1)
    fig.update_yaxes(title_text="Thickness (m)", row=2, col=1)
    fig.update_xaxes(title_text="Radial Station", row=2, col=2)
    fig.update_yaxes(title_text="Mid Chord Alignment (m)", row=2, col=2)

    fig.update_layout(title_text="Propeller Geometry", height=700, showlegend=False)

    if save_figure:
        fig.write_image(save_filename + '_2D' + file_type)
    
    if show_figure:
        fig.write_html( save_filename + '.html', auto_open=True)

    return fig



# RCAIDE/Framework/Analyses/Mission/Segments/Ground/Battery_Charge.py 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
from RCAIDE.Framework.Mission.Segments.Evaluate      import Evaluate
from RCAIDE.Library.Mission.Segments                 import Ground   
from RCAIDE.Framework.Core                           import Units
from RCAIDE.Library.Methods.skip                     import skip 

# ----------------------------------------------------------------------------------------------------------------------
#  SEGMENT
# ---------------------------------------------------------------------------------------------------------------------- 
class Battery_Recharge(Evaluate): 

    # ------------------------------------------------------------------
    #   Data Defaults
    # ------------------------------------------------------------------  

    def __defaults__(self):  

        """ This sets the default solver flow. Anything in here can be modified after initializing a segment.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """              
        
        # --------------------------------------------------------------
        #   User Inputs
        # --------------------------------------------------------------
        self.altitude                      = None 
        self.overcharge_contingency        = 1.10
        self.cutoff_SOC                    = 1.0
        self.true_course                   = 0.0 * Units.degrees  
        self.cooling_time                  = 0.0 * Units.hr
         
        initialize                         = self.process.initialize  
        initialize.conditions              = Ground.Battery_Charge_Discharge.initialize_conditions  
        iterate                            = self.process.iterate 
        iterate.unknowns.mission           = skip
        iterate.conditions.aerodynamics    = skip
        iterate.conditions.stability       = skip  
        post_process                       = self.process.post_process  
        post_process.noise                 = skip
        post_process.emissions             = skip
        
        return



## @ingroup Methods-Propulsion-Rotor_Wake-Fidelity_One
# compute_fidelity_one_inflow_velocities.py
#
# Created:  Sep 2021, R. Erhard
# Modified: Jan 2022, R. Erhard

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------
from RCAIDE.Framework.Core import Data
from RCAIDE.Library.Methods.Propulsors.Converters.Rotor.Wake.Perscribed_Vortex_Wake import compute_wake_induced_velocity

# package imports
import numpy as np
from scipy.interpolate import interp1d

## @ingroup Methods-Propulsion-Rotor_Wake-Fidelity_One
def compute_fidelity_one_inflow_velocities( wake, prop ):
    """
    Assumptions:
        None

    Source:
        N/A
    Inputs:
        wake - rotor wake
        prop - rotor instance
        WD   - wake vortex distribution
    Outputs:
        Va   - axial velocity, shape (ctrl_pts, Nr, Na); axis 2 in direction of rotation       [m/s]
        Vt   - tangential velocity, shape (ctrl_pts, Nr, Na); axis 2 in direction of rotation    [m/s]
    """
    
    VD                       = prop.vortex_distribution
    omega                    = prop.inputs.omega
    init_timestep_offset     = wake.wake_settings.initial_timestep_offset

    # use results from prior bevw iteration
    prop_outputs  = prop.outputs
    cpts          = len(prop_outputs.velocity)
    Na            = prop.number_azimuthal_stations
    Nr            = len(prop.chord_distribution)
    r             = prop.radius_distribution
    rot           = prop.rotation
    WD            = wake.vortex_distribution


    try:
        props = prop.propellers_in_network
    except:
        props = Data()
        props.propeller = prop

    # compute radial blade section locations based on initial timestep offset
    azi_step = 2*np.pi/(Na+1)
    dt       = azi_step/omega[0][0]
    t0       = dt*init_timestep_offset

    # set shape of velocitie arrays
    Va = np.zeros((cpts,Nr,Na))
    Vt = np.zeros((cpts,Nr,Na))
    
    for i in range(Na):
        # increment blade angle to new azimuthal position 
        blade_angle   = -rot*(omega[0]*t0 + i*(2*np.pi/(Na)))  # axial view of rotor, negative rotation --> positive blade angle
    
        #----------------------------------------------------------------
        #Compute the wake-induced velocities at propeller blade
        #----------------------------------------------------------------
        #set the evaluation points in the vortex distribution: (ncpts, nblades, Nr, Ntsteps)
        r    = prop.radius_distribution 
        Yb   = wake.vortex_distribution.reshaped_wake.Yblades_cp[i,0,0,:,0]
        Zb   = wake.vortex_distribution.reshaped_wake.Zblades_cp[i,0,0,:,0]
        Xb   = wake.vortex_distribution.reshaped_wake.Xblades_cp[i,0,0,:,0]
        
        VD.YC = (Yb[1:] + Yb[:-1])/2
        VD.ZC = (Zb[1:] + Zb[:-1])/2
        VD.XC = (Xb[1:] + Xb[:-1])/2
         
        VD.n_cp = np.size(VD.YC)

        # Compute induced velocities at blade from the helical fixed wake
        VD.Wake_collapsed = WD
        
        V_ind   = compute_wake_induced_velocity(WD, VD, cpts, azi_start_idx=i)
        
        # velocities in vehicle frame
        u       = V_ind[:,:,0]   # velocity in vehicle x-frame
        v       = V_ind[:,:,1]    # velocity in vehicle y-frame
        w       = V_ind[:,:,2]    # velocity in vehicle z-frame
        
        # rotate from vehicle to prop frame:
        rot_to_prop = prop.vec_to_prop_body()
        uprop       = u*rot_to_prop[:,0,0][:,None] + w*rot_to_prop[:,0,2][:,None]
        vprop       = v
        wprop       = u*rot_to_prop[:,2,0][:,None] + w*rot_to_prop[:,2,2][:,None]     
        
        # interpolate to get values at rotor radial stations
        r_midpts = (r[1:] + r[:-1])/2
        u_r      = interp1d(r_midpts, uprop, fill_value="extrapolate")
        v_r      = interp1d(r_midpts, vprop, fill_value="extrapolate")
        w_r      = interp1d(r_midpts, wprop, fill_value="extrapolate")
        
        up = u_r(r)
        vp = v_r(r)
        wp = w_r(r)       

        # Update velocities at the disc
        Va[:,:,i]  = up
        Vt[:,:,i]  = -rot*(vp*(np.cos(blade_angle)) - wp*(np.sin(blade_angle)) )  # velocity component in direction of rotation     
    
    prop.vortex_distribution = VD
    

    return Va, Vt




# ContainerOrdered.py
#
# Created:  Jan 2015, T. Lukacyzk
# Modified: Feb 2016, T. MacDonald
#           Jun 2016, E. Botero
#           May 2020, E. Botero



# ----------------------------------------------------------------------
#   Imports
# ----------------------------------------------------------------------        

from .DataOrdered import DataOrdered

# ----------------------------------------------------------------------
#   Data Container Base Class
# ----------------------------------------------------------------------        

class ContainerOrdered(DataOrdered):
    """ A dict-type container with attribute, item and index style access
        intended to hold a attribute-accessible list of DataOrdered(). This is ordered.
        
        Assumptions:
        N/A
        
        Source:
        N/A
        
    """
        
    def __defaults__(self):
        """Defaults function
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A
            """          
        pass
    
    def __init__(self,*args,**kwarg):
        """Initialization that builds the container
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            self
    
            Outputs:
            N/A
    
            Properties Used:
            N/A
            """           
        super(ContainerOrdered,self).__init__(*args,**kwarg)
        self.__defaults__()
    
    def append(self,val):
        """Appends the value to the containers
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            self
    
            Outputs:
            N/A
    
            Properties Used:
            N/A
            """          
        #val = self.check_new_val(val)
        DataOrdered.append(self,val)
        
    def get_children(self):
        """ Returns the components that can go inside
        
        Assumptions:
        None
    
        Source:
        N/A
    
        Inputs:
        None
    
        Outputs:
        None
    
        Properties Used:
        N/A
        """        
        
        return []  


# RCAIDE/Library/Plots/Thermal_Management/__init__.py


"""RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .plot_thermal_management_performance             import plot_thermal_management_performance
from .plot_wavy_channel_conditions                   import plot_wavy_channel_conditions
from .plot_cross_flow_heat_exchanger_conditions      import plot_cross_flow_heat_exchanger_conditions
from .plot_reservoir_conditions                      import plot_reservoir_conditions
from .plot_air_cooled_conditions                     import plot_air_cooled_conditions


#  RCAIDE/Methods/Energy/Distributors/Electrical_Bus/initialize_bus_properties.py
# 
# Created: Sep 2024, S. Shekar
#
from RCAIDE.Library.Methods.Energy.Sources.Batteries.Common                    import compute_module_properties 

# ----------------------------------------------------------------------------------------------------------------------
#  METHODS
# ---------------------------------------------------------------------------------------------------------------------- 
def initialize_bus_properties(bus): 
    """ Initializes the bus electrical properties based what is appended onto the bus
        
        Assumptions:
        N/A
    
        Source:
        N/A
    
        Inputs:  
       
        Outputs:
           
        Properties Used:
        None
        """
    if bus.battery_module_electric_configuration == 'Series':
        bus.nominal_capacity = 0
        bus.maximum_energy   = 0
        for battery_module in  bus.battery_modules: 
            compute_module_properties(battery_module) 
            bus.voltage         +=  battery_module.voltage
            bus.maximum_energy  +=  battery_module.maximum_energy
            bus.nominal_capacity =  max(battery_module.nominal_capacity, bus.nominal_capacity)  
    elif bus.battery_module_electric_configuration == 'Parallel':
        bus.voltage = 0
        bus.maximum_energy   = 0
        for battery_module in  bus.battery_modules: 
            compute_module_properties(battery_module)        
            bus.voltage           =  max(battery_module.voltage, bus.voltage)
            bus.nominal_capacity +=  battery_module.nominal_capacity        
            bus.maximum_energy  +=  battery_module.initial_maximum_energy 
    return


# RCAIDE/Methods/Noise/Common/generate_microphone_locations.py
# 
# 
# Created:  Oct 2023, A. Molloy  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports  
from RCAIDE.Framework.Core import Units, Data
from RCAIDE.Framework.Analyses.Geodesics.Geodesics import Calculate_Distance

# package imports 
from scipy.interpolate import griddata
import numpy as np 
 
# ---------------------------------------------------------------------------------------------------------------------- 
#  generate_terrain_microphone_locations
# ----------------------------------------------------------------------------------------------------------------------  
def generate_terrain_microphone_locations(settings):
    """This computes the absolute microphone/observer locations on a defined topography
            
    Assumptions: 
        topography_file is a text file obtained from https://topex.ucsd.edu/cgi-bin/get_data.cgi
    
    Source:
        N/A  

    Inputs:  
        topography_file                        - file of lattide, longitude and elevation points                                [-]
        origin_coordinates                  - coordinates of origin location                                              [degrees]
        destination_coordinates                - coordinates of destimation location                                            [degrees]
        microphone_x_resolution         - number of points on computational domain in latitudal direction                [-]
        microphone_y_resolution         - number of points on computational domain in  longitidinal direction            [-] 
        ground_microphone_x_stencil            - number of points in stencil in latitudal direction                             [-] 
        adjusted_cruise_distance                       - distance used to modify cruise to ensure desired range is met                  [-]
        ground_microphone_y_stencil            - number of points in stencil in in longitidinal direction                       [-] 
        
    Outputs: 
    topography_data.
        microphone_x_resolution         - number of points on computational domain in latitudal direction                [-] 
        microphone_y_resolution         - number of points on computational domain in  longitidinal direction            [-]
        ground_microphone_x_stencil            - number of points in stencil in latitudal direction                             [-]
        ground_microphone_y_stencil            - number of points in stencil in in longitidinal direction                       [-]       
        microphone_min_x                - x-location of start of computation domain                                      [meters]          
        microphone_max_x                - x-location of end of computation domain                                        [meters]  
        microphone_min_y                - y-location of start of computation domain                                      [meters]                
        microphone_max_y                - y-location of end of computation domain                                        [meters]  
        cartesian_micrphone_locations          - cartesian coordinates (x,y,z) of all microphones in domain                     [meters]       
        latitude_longitude_micrphone_locations - latitude-longitude and elevation coordinates of all microphones in domain      [deg,deg,m]   
    
    Properties Used:
        N/A       
    """     
    
    y_res = settings.microphone_y_resolution 
    x_res = settings.microphone_x_resolution 
    
    # extract data from file 
    data  = np.loadtxt(settings.topography_file) # settings.topography_file) CHANGED 10-15-2024
    Long  = data[:,0]
    Lat   = data[:,1]
    Elev  = data[:,2] 
    
    x_min_coord = np.min(Lat)
    x_max_coord = np.max(Lat)
    y_min_coord = np.min(Long)
    y_max_coord = np.max(Long) 

    if y_min_coord < 0: 
        y_min_coord = 360 + y_min_coord
    if y_max_coord< 0:
        y_max_coord=360 + y_max_coord 
    
    top_left_map_coords      = np.array([x_max_coord,y_min_coord])
    bottom_left_map_coords   = np.array([x_min_coord,y_min_coord])  
    bottom_right_map_coords  = np.array([x_min_coord,y_max_coord]) 
    
    x_dist_max = Calculate_Distance(top_left_map_coords,bottom_left_map_coords) * Units.kilometers
    y_dist_max = Calculate_Distance(bottom_right_map_coords,bottom_left_map_coords) * Units.kilometers
    
    [y_pts,x_pts]      = np.meshgrid(np.linspace(0,y_dist_max,y_res),np.linspace(0,x_dist_max,x_res))
    [long_deg,lat_deg] = np.meshgrid(np.linspace(np.min(Long),np.max(Long),y_res),np.linspace(np.min(Lat),np.max(Lat),x_res)) 
    z_deg              = griddata((Lat,Long), Elev, (lat_deg, long_deg), method='linear')        
    cartesian_pts      = np.dstack((np.dstack((x_pts[:,:,None],y_pts[:,:,None] )),z_deg[:,:,None])).reshape(x_res*y_res,3)
    lat_long_pts       = np.dstack((np.dstack((lat_deg[:,:,None],long_deg[:,:,None] )),z_deg[:,:,None])).reshape(x_res*y_res,3)  
    return cartesian_pts , lat_long_pts





# RCAIDE/Library/Methods/Aerodynamics/Athena_Vortex_Lattice/AVL_Objects/Configuration.py
#
# Created: Oct 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
from   RCAIDE.Framework.Core import  Data  

# ----------------------------------------------------------------------------------------------------------------------
# Inputs
# ---------------------------------------------------------------------------------------------------------------------- 
class Configuration(Data):
    """A data class defining the reference parameters of the aircraft geometry and 
    flight configuration 

    Assumptions:
        None

    Source:
        None

    Inputs:
        None

    Outputs:
        None

    Properties Used:
        N/A
    """    


    def __defaults__(self):
        """ Defines the data structure and defaults for mass properties of the aircraft 
        in AVL

        Assumptions:
            None

        Source:
            None

        Inputs:
            None

        Outputs:
            None

        Properties Used:
            N/A
        """ 		
        self.tag                          = 'configuration'
        self.parasite_drag                = 0.0

        self.reference_values             = Data()     
        self.reference_values.sref        = 0.0        # [m]
        self.reference_values.bref        = 0.0        # [m]
        self.reference_values.cref        = 0.0        # [m]
        self.reference_values.cg_coords   = [0.,0.,0.] # [m]

        self.mass_properties              = Data()
        self.mass_properties.inertial     = Data()
        self.mass_properties.mass         = 0.0        # [kg]
        self.mass_properties.inertial.Ixx = 0.0        # [kg.m^2]   
        self.mass_properties.inertial.Iyy = 0.0        # [kg.m^2]   
        self.mass_properties.inertial.Izz = 0.0        # [kg.m^2]   
        self.mass_properties.inertial.Ixy = 0.0        # [kg.m^2]   
        self.mass_properties.inertial.Iyz = 0.0        # [kg.m^2]   
        self.mass_properties.inertial.Izx = 0.0        # [kg.m^2]

        self.symmetry_settings            = Data()
        self.symmetry_settings.Iysym      = 0	# Assumed y-symmetry of solution  # (1: symmetric, -1: antisymmetric, 0: no symmetry assumed)
        self.symmetry_settings.Izsym      = 0	# Assumed z-symmetry of solution
        self.symmetry_settings.Zsym       = 0.0	# z-coordinate of plane of z-symmetry


# RCAIDE/Library/Plots/Performance/Aerodynamics/plot_rotor_conditions.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style 
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------      
def plot_rotor_conditions(results,
                        save_figure = False,
                        show_legend=True,
                        save_filename = "Rotor_Conditions",
                        file_type = ".png",
                        width = 11, height = 7):
    """This plots the electric driven network propeller efficiencies 

    Assumptions:
    None

    Source:
    None

    Inputs:
    results.segments.conditions.propulsion.  
        
    Outputs: 
    Plots

    Properties Used:
    N/A	
    """	   
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters) 
    
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))   

    fig = plt.figure(save_filename)
    fig.set_size_inches(width,height)  
    axis_1 = plt.subplot(2,2,1)
    axis_2 = plt.subplot(2,2,2)
    axis_3 = plt.subplot(2,2,3) 
    axis_4 = plt.subplot(2,2,4)   
 
    for network in results.segments[0].analyses.energy.vehicle.networks: 
        for p_i, propulsor in enumerate(network.propulsors): 
            if (p_i == 0) or (network.identical_propulsors == False):            
                plot_propulsor_data(results,propulsor,axis_1,axis_2,axis_3,axis_4,line_colors,ps,p_i)                  
              
    if show_legend:                
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4) 
        leg.set_title('Flight Segment', prop={'size': ps.legend_font_size, 'weight': 'heavy'})    
    
    # Adjusting the sub-plots for legend
    fig.tight_layout() 
    fig.subplots_adjust(top=0.8) 
    
    # set title of plot 
    title_text  =  'Rotor Performance' 
    fig.suptitle(title_text)
    if save_figure:
        plt.savefig(save_filename + file_type) 
                 
    return fig 

def plot_propulsor_data(results,propulsor,axis_1,axis_2,axis_3,axis_4,line_colors,ps,p_i):
    if 'rotor' in  propulsor:
        thrustor = propulsor.rotor
    elif 'propeller' in  propulsor:
        thrustor = propulsor.propeller
        
    for i in range(len(results.segments)): 
        bus_results  =  results.segments[i].conditions.energy
        time         =  results.segments[i].conditions.frames.inertial.time[:,0] / Units.min   
        rpm          =  bus_results[propulsor.tag][thrustor.tag].rpm[:,0]
        thrust       =  np.linalg.norm(bus_results[propulsor.tag][thrustor.tag].thrust , axis =1)
        torque       =  bus_results[propulsor.tag][thrustor.tag].torque[:,0]
        angle        =  bus_results[propulsor.tag].commanded_thrust_vector_angle[:,0]   
        if  i == 0 :
            axis_1.plot(time,rpm, color = line_colors[i], marker = ps.markers[p_i]  , linewidth = ps.line_width, label = thrustor.tag)
        else:
            axis_1.plot(time,rpm, color = line_colors[i], marker = ps.markers[p_i]  , linewidth = ps.line_width)
        axis_1.set_ylabel(r'RPM')
        set_axes(axis_1)    
         
        axis_2.plot(time, angle/Units.degrees, color = line_colors[i], marker = ps.markers[p_i]  , linewidth = ps.line_width) 
        axis_2.set_ylabel(r'Rotor Angle')
        set_axes(axis_2) 
 
        axis_3.plot(time,thrust, color = line_colors[i], marker = ps.markers[p_i] , linewidth = ps.line_width)
        axis_3.set_ylabel(r'Thrust (N)')
        set_axes(axis_3) 
         
        axis_4.plot(time,torque, color = line_colors[i], marker = ps.markers[p_i] , linewidth = ps.line_width)
        axis_4.set_ylabel(r'Torque (N-m)')
        set_axes(axis_4)     
    return 


# RCAIDE/Library/Methods/Thermal_Management/Heat_Exchangers/Cross_Flow_Heat_Exchanger/design_cross_flow_heat_exchanger.py
#
# Created:  Apr 2024, S. Shekar 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports 
from RCAIDE.Framework.Core                            import Units, Data   
from RCAIDE.Library.Methods.Thermal_Management.Heat_Exchangers.Cross_Flow_Heat_Exchanger.cross_flow_heat_exchanger_sizing_setup import cross_flow_heat_exchanger_sizing_setup 
from RCAIDE.Library.Methods.Thermal_Management.Heat_Exchangers.Cross_Flow_Heat_Exchanger.cross_flow_heat_exchanger_geometry_setup import cross_flow_heat_exchanger_geometry_setup
from RCAIDE.Framework.Optimization.Common             import Nexus
from RCAIDE.Framework.Optimization.Packages.scipy     import scipy_setup

# Python package imports   
import numpy as np  
import time 

# ----------------------------------------------------------------------
#  Rotor Design
# ---------------------------------------------------------------------- 
def design_cross_flow_heat_exchanger(HEX,coolant_line,battery, single_side_contact=True, dry_mass=True,
                                       solver_name= 'SLSQP',iterations = 200,solver_sense_step = 1E-3,
                                       solver_tolerance = 1E-2,print_iterations = False):
    
    """ Optimizes heat exchanger geometric properties input parameters to minimize either design power and mass. 
        This scrip adopts RCAIDE's native optimization style where the objective function is expressed 
        as a sizing function, considering both power and mass.
          
          Assumptions: 
            
        
          Source:
           Shah RK, Sekulić DP. Fundamentals of Heat Exchanger Design. John Wiley & Sons; 2003. 
    """    
    
    
    # start optimization 
    ti                   = time.time()  
    optimization_problem = crossflow_heat_exchanger_design_problem_setup(HEX,coolant_line,print_iterations)
    output               = scipy_setup.SciPy_Solve(optimization_problem,solver=solver_name, iter = iterations , sense_step = solver_sense_step,tolerance = solver_tolerance)  
    
    tf                   = time.time()
    elapsed_time         = round((tf-ti)/60,2)
    print('HEX sizing optimization Simulation Time: ' + str(elapsed_time) + ' mins')    
 

    for coolant_line in optimization_problem.hex_configurations.optimized.networks.electric.coolant_lines:
        for heat_exchanger in coolant_line.heat_exchangers: 
            HEX_opt       = heat_exchanger
             
    HEX.design_air_inlet_pressure     = output[0]
    HEX.design_coolant_inlet_pressure = output[1]
    HEX.design_air_mass_flow_rate     = output[2]
    HEX.design_coolant_mass_flow_rate = output[3]
    HEX.stack_length                  = HEX_opt.stack_length 
    HEX.stack_width                   = HEX_opt.stack_width  
    HEX.stack_height                  = HEX_opt.stack_height  
    HEX.design_air_frontal_area       = HEX_opt.air_frontal_area
    HEX.design_air_pressure_diff      = HEX_opt.pressure_diff_air
    HEX.design_heat_removed           = HEX_opt.heat_removed
    HEX.mass_properties.mass          = HEX_opt.heat_exchanger_mass
    return HEX
 
def crossflow_heat_exchanger_design_problem_setup(HEX,coolant_line,print_iterations): 
    """ Sets up cross flow heat exchanger optimization problem including design variables,
        constraints and objective function using RCAIDE's Nexus optimization framework.
        Appends methodolody of planform modification to Nexus.
          Inputs:
             HEX     - heat exchanger data structure            [None]
             coolant_line     - Coolant Line data structure           [None]
             
          Outputs: 
              nexus    - RCAIDE's optimization framework [None]
              
          Assumptions:
            None
        
          Source:
             None
    """        
   
    nexus                        = Nexus()
    problem                      = Data()
    nexus.optimization_problem   = problem    
    
    
    # ---------------------------------------------------------------------------------------------------------- 
    # Design Variables 
    # ----------------------------------------------------------------------------------------------------------       
    inputs = []   
    #               variable   initial   lower limit   upper limit     scaling       units 
    inputs.append([ 'p_c_1'   ,   200e3     ,  200e3     , 220e3+1       , 1     ,  1*Units.less]) 
    inputs.append([ 'p_h_1'   ,   160e3     ,  160e3     , 180e3+1       , 1     ,  1*Units.less]) 
    inputs.append([ 'm_dot_c' ,   1         ,  0.1       , 5          , 1.0       ,  1*Units.less])  
    inputs.append([ 'm_dot_h' ,   0.1       ,  0.1       , 1             , 1.0       ,  1*Units.less])  
   
        
    problem.inputs = np.array(inputs,dtype=object)   

    # ----------------------------------------------------------------------------------------------------------
    # Objective
    # ---------------------------------------------------------------------------------------------------------- 
    problem.objective = np.array([  
                                 [  'P_hex'  ,  100000   ,    1*Units.less] 
    ],dtype=object)
            

    # ----------------------------------------------------------------------------------------------------------
    # Constraints
    # ----------------------------------------------------------------------------------------------------------  
    constraints = []      
    constraints.append([ 'L'         ,  '<'  ,  2.0 ,   1.0   , 1*Units.less])
    constraints.append([ 'W'         ,  '<'  ,  3.0 ,   1.0   , 1*Units.less])  
    constraints.append([ 'H'         ,  '<'  ,  2.0 ,   1.0   , 1*Units.less]) 
    constraints.append([ 'L'         ,  '>'  ,  0.1 ,   1.0   , 1*Units.less])
    constraints.append([ 'W'         ,  '>'  ,  0.1 ,   1.0   , 1*Units.less])  
    constraints.append([ 'H'         ,  '>'  ,  0.1 ,   1.0   , 1*Units.less])
    constraints.append([ 'Q_res'     ,  '<'  ,  5,   1.0   , 1*Units.less])
    problem.constraints =  np.array(constraints,dtype=object)                
    
    # -------------------------------------------------------------------
    #  Aliases
    # ------------------------------------------------------------------- 
    aliases = [] 
    btms = 'hex_configurations.optimized.networks.electric.coolant_lines.' +  coolant_line.tag + '.heat_exchangers.'+  HEX.tag
    aliases.append([ 'm_dot_c'     , btms + '.air_flow_rate' ]) 
    aliases.append([ 'm_dot_c'     , btms + '.coolant_flow_rate' ]) 
    aliases.append([ 'p_c_1'       , btms + '.air_inlet_pressure' ])       
    aliases.append([ 'p_h_1'       , btms + '.coolant_inlet_pressure' ])   
    aliases.append([ 'L'           , 'summary.stack_length' ])     
    aliases.append([ 'W'           , 'summary.stack_width' ])   
    aliases.append([ 'H'           , 'summary.stack_height' ])        
    aliases.append([ 'P_hex'       , 'summary.mass_power_objective' ])
    aliases.append([ 'Q_res'       , 'summary.heat_residual' ])   
    problem.aliases = aliases
    
    # -------------------------------------------------------------------
    #  Vehicles
    # ------------------------------------------------------------------- 
    nexus.hex_configurations = cross_flow_heat_exchanger_geometry_setup(HEX,coolant_line) 
    # -------------------------------------------------------------------
    #  Analyses
    # -------------------------------------------------------------------
    nexus.analyses = None 
    
    # -------------------------------------------------------------------
    #  Missions
    # -------------------------------------------------------------------
    nexus.missions = None
    
    # -------------------------------------------------------------------
    #  Procedure
    # -------------------------------------------------------------------    
    nexus.print_iterations  = print_iterations 
    nexus.procedure         = cross_flow_heat_exchanger_sizing_setup()
    
    # -------------------------------------------------------------------
    #  Summary
    # -------------------------------------------------------------------    
    nexus.summary         = Data()     
         
    return nexus   




# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# package imports
import numpy as np


# ----------------------------------------------------------------------
#  Compressible Mixed Flat Plate
# ----------------------------------------------------------------------

## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Helper_Functions
def compressible_mixed_flat_plate(Re,Ma,Tc,xt):
    """Computes the coefficient of friction for a flat plate given the 
    input parameters. Also returns the correction terms used in the
    computation.

    Assumptions:
    Reynolds number between 10e5 and 10e9
    xt between 0 and 1

    Source:
    adg.stanford.edu (Stanford AA241 A/B Course Notes)

    Inputs:
    Re (Reynolds number)                                             [Unitless]
    Ma (Mach number)                                                 [Unitless]
    Tc (temperature)                                                 [K]
    xt (turbulent transition point as a proportion of chord length)  [Unitless]

    Outputs:
    cf_comp (coefficient of friction)                                [Unitless]
    k_comp (compressibility correction)                              [Unitless]
    k_reyn (Reynolds number correction)                              [Unitless]

    Properties Used:
    N/A
    """     
    
    if xt < 0.0 or xt > 1.0:
        raise ValueError("Turbulent transition must be between 0 and 1")
    
    #if np.any(Re > 10**9) or np.any(Re < 10**5):
        ##print 'Warning: Reynolds number outside expected range - in file compressible_mixed_flat_plate.py'
        #pass
    
    Rex = Re*xt
    Rex[Rex==0.0] = 0.0001

    theta = 0.671*xt/(Rex**0.5)
    xeff  = (27.78*theta*Re**0.2)**1.25
    Rext  = Re*(1-xt+xeff)
    
    cf_turb  = 0.455/(np.log10(Rext)**2.58)
    cf_lam   = 1.328/(Rex**0.5)
    
    if xt > 0.0:
        cf_start = 0.455/(np.log10(Re*xeff)**2.58)
    else:
        cf_start = 0.0
    
    cf_inc = cf_lam*xt + cf_turb*(1-xt+xeff) - cf_start*xeff
    
    # compressibility correction
    Tw = Tc * (1. + 0.178*Ma*Ma)
    Td = Tc * (1. + 0.035*Ma*Ma + 0.45*(Tw/Tc - 1.))
    k_comp = (Tc/Td) 
    
    # reynolds correction
    Rd_w   = Re * (Td/Tc)**1.5 * ( (Td+216.) / (Tc+216.) )
    k_reyn = (Re/Rd_w)**0.2
    
    # apply corrections
    cf_comp = cf_inc * k_comp * k_reyn
    
    return cf_comp, k_comp, k_reyn

  
# ----------------------------------------------------------------------
#   Module Tests
# ----------------------------------------------------------------------
# this will run from command line, put simple tests for your code here
if __name__ == '__main__':    
    
    validation_test = 0
    import pylab as plt
    
    if validation_test:
    
        Re = np.logspace(5,9,5)
        ii = 0
        xts = np.array([0.0,.1,.2,.3,.4,.5,.6,.8,1.0])
        cf_comp = np.zeros([9,5])
        k_comp = np.zeros_like(cf_comp)
        k_reyn = np.zeros_like(cf_comp)
        for xt in xts:
            (cf_comp[ii,:], k_comp[ii,:], k_reyn[ii,:]) = compressible_mixed_flat_plate(Re,0.0,216.0,xt)
            if ii == 0:
                (cf_comp_turb,k_comp_t,k_reyn_t) = compressible_turbulent_flat_plate(Re,0.0,216.0)
            ii = ii + 1
            
        cf_comp_empirical = np.array([[.0073,.0045,.0031,.0022,.0016],
                                      [.0074,.0044,.0029,.0020,.0015],
                                      [.0073,.0041,.0026,.0018,.0014],
                                      [.0070,.0038,.0024,.0017,.0013],
                                      [.0067,.0036,.0021,.0015,.0011],
                                      [.0064,.0033,.0020,.0013,.0009],
                                      [.0060,.0029,.0017,.0011,.0007],
                                      [.0052,.0022,.0011,.0006,.0004],
                                      [.0041,.0013,.0004,.0001,.00005]])
        
            
        plt.figure("Skin Friction Coefficient v. Reynolds Number")
        axes = plt.gca()    
        for i in range(len(cf_comp[:,0])):
            axes.semilogx(Re, cf_comp_empirical[i,:], 'ro-')
            axes.semilogx(Re, cf_comp[i,:], 'bo-')
            axes.semilogx(Re, cf_comp_turb, 'go-')
        axes.set_xlabel('Reynolds Number')
        axes.set_ylabel('Cf')
        axes.grid(True)  
        plt.show()

    else:
        
        Re = 10**7
        Ma = 2.0
        Tc = 216.0
        xt = 0.6
        (cf_comp, k_comp, k_reyn) = compressible_mixed_flat_plate(Re, Ma, Tc, xt)
        
    print(cf_comp)
    print(k_comp)
    print(k_reyn)    


# RCAIDE/Library/Components/__init__.py

"""
Module containing core aircraft component classes and submodules for aircraft design 
and analysis.

See Also
--------
RCAIDE.Library.Components.Component
    Base component class
RCAIDE.Library.Components.Network
    Network component class for system connections
RCAIDE.Library.Components.Mass_Properties
    Mass properties data structure
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from .Component        import Component
from . Network         import Network
from .Mass_Properties  import Mass_Properties
  
from . import Propulsors
from . import Energy
from . import Thermal_Management 
from . import Airfoils
from . import Booms
from . import Configs
from . import Fuselages
from . import Landing_Gear
from . import Nacelles
from . import Payloads
from . import Systems
from . import Wings